import warnings
from typing import Any, Iterable, Literal, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.collections import PathCollection, PolyCollection
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure, SubFigure
from matplotlib.lines import Line2D
from matplotlib.offsetbox import AnchoredOffsetbox, AnchoredText
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import (
    ALONG_TRACK_DIM,
    ELEVATION_VAR,
    FIGURE_HEIGHT_CURTAIN,
    FIGURE_HEIGHT_LINE,
    FIGURE_WIDTH_CURTAIN,
    FIGURE_WIDTH_LINE,
    HEIGHT_VAR,
    TEMP_CELSIUS_VAR,
    TIME_VAR,
    TRACK_LAT_VAR,
    TRACK_LON_VAR,
    TROPOPAUSE_VAR,
)
from ...utils.ground_sites import GroundSite, get_ground_site
from ...utils.overpass import get_overpass_info
from ...utils.profile_data import (
    ProfileData,
    ensure_along_track_2d,
    ensure_vertical_2d,
    validate_profile_data_dimensions,
)
from ...utils.time import (
    TimedeltaLike,
    TimeRangeLike,
    TimestampLike,
    to_timedelta,
    to_timestamp,
    to_timestamps,
    validate_time_range,
)
from ...utils.typing import DistanceRangeLike, ValueRangeLike
from ..color import Cmap, Color, ColorLike, get_cmap
from ..save import save_plot
from ._plot_1d_integer_flag import plot_1d_integer_flag
from ._plot_stacked_propabilities import plot_stacked_propabilities
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .annotation import (
    add_text,
    add_text_product_info,
    add_title_earthcare_frame,
    format_var_label,
)
from .colorbar import add_colorbar
from .defaults import get_default_cmap, get_default_norm, get_default_rolling_mean
from .height_ticks import format_height_ticks
from .ticks import format_numeric_ticks
from .value_range import select_value_range


class LineFigure:
    """TODO: documentation"""

    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (FIGURE_WIDTH_LINE, FIGURE_HEIGHT_LINE),
        dpi: int | None = None,
        title: str | None = None,
        ax_style_top: AlongTrackAxisStyle | str = "geo",
        ax_style_bottom: AlongTrackAxisStyle | str = "time",
        num_ticks: int = 10,
        show_value_left: bool = True,
        show_value_right: bool = False,
        mode: str | Literal["line", "scatter", "area"] = "line",
        show_grid: bool = True,
        grid_color: str | None = Color("lightgray"),
        grid_which: Literal["major", "minor", "both"] = "major",
        grid_axis: Literal["both", "x", "y"] = "both",
        grid_linestyle: str = "dashed",
        grid_linewidth: float = 1,
        fig_height_scale: float = 1.0,
        fig_width_scale: float = 1.0,
    ):
        figsize = (figsize[0] * fig_width_scale, figsize[1] * fig_height_scale)
        self.fig: Figure
        if isinstance(ax, Axes):
            tmp = ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = ax
        else:
            self.fig = plt.figure(figsize=figsize, dpi=dpi)
            self.ax = self.fig.add_subplot()  # add_axes((0.0, 0.0, 1.0, 1.0))
        self.title = title
        if self.title:
            self.fig.suptitle(self.title, y=1.4)

        self.ax_top: Axes
        self.ax_right: Axes
        self.selection_time_range: tuple[pd.Timestamp, pd.Timestamp] | None = None
        self.ax_style_top: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_top
        )
        self.ax_style_bottom: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_bottom
        )

        self.info_text: AnchoredText | None = None
        self.info_text_loc: str = "upper right"
        self.num_ticks = num_ticks
        self.show_value_left: bool = show_value_left
        self.show_value_right: bool = show_value_right
        self.mode: str | Literal["line", "scatter", "area"] = mode

        self.show_grid = show_grid
        self.grid_color = Color.from_optional(grid_color)
        self.grid_which = grid_which
        self.grid_axis = grid_axis
        self.grid_linestyle = grid_linestyle
        self.grid_linewidth = grid_linewidth

        self.ax_right = self.ax.twinx()
        self.ax_right.set_ylim(self.ax.get_ylim())
        self.ax_right.set_yticks([])

        self.ax_top = self.ax.twiny()
        self.ax_top.set_xlim(self.ax.get_xlim())

        self.tmin: np.datetime64 | None = None
        self.tmax: np.datetime64 | None = None

    def _set_info_text_loc(self, info_text_loc: str | None) -> None:
        if isinstance(info_text_loc, str):
            self.info_text_loc = info_text_loc

    def _set_axes(
        self,
        tmin: np.datetime64,
        tmax: np.datetime64,
        vmin: float,
        vmax: float,
        time: NDArray,
        tmin_original: np.datetime64 | None = None,
        tmax_original: np.datetime64 | None = None,
        longitude: NDArray | None = None,
        latitude: NDArray | None = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
    ) -> "LineFigure":
        if ax_style_top is not None:
            self.ax_style_top = AlongTrackAxisStyle.from_input(self.ax_style_top)
        if ax_style_bottom is not None:
            self.ax_style_bottom = AlongTrackAxisStyle.from_input(self.ax_style_bottom)
        if not isinstance(tmin_original, np.datetime64):
            tmin_original = tmin
        if not isinstance(tmax_original, np.datetime64):
            tmax_original = tmax

        self.ax.set_xlim((tmin, tmax))  # type: ignore
        self.ax.set_ylim((vmin, vmax))

        if self.show_grid:
            self.ax.grid(
                visible=self.show_grid,
                which=self.grid_which,
                axis=self.grid_axis,
                color=self.grid_color,
                linestyle=self.grid_linestyle,
                linewidth=self.grid_linewidth,
            )

        self.ax_right.set_ylim(self.ax.get_ylim())
        self.ax_top.set_xlim(self.ax.get_xlim())

        format_along_track_axis(
            self.ax,
            self.ax_style_bottom,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude,
            latitude,
            num_ticks=self.num_ticks,
        )
        format_along_track_axis(
            self.ax_top,
            self.ax_style_top,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude,
            latitude,
            num_ticks=self.num_ticks,
        )
        return self

    def plot(
        self,
        *,
        values: NDArray | None = None,
        time: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        # Common args for wrappers
        mode: str | Literal["line", "scatter", "area"] | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        color: str | None = Color("ec:blue"),
        alpha: float = 1.0,
        linestyle: str | None = "solid",
        linewidth: float | int | None = 2.0,
        marker: str | None = "s",
        markersize: float | int | None = 2.0,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str | None = Color("white"),
        selection_highlight_alpha: float = 0.5,
        selection_max_time_margin: (
            TimedeltaLike | Sequence[TimedeltaLike] | None
        ) = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        mark_profiles_at: Sequence[TimestampLike] | None = None,
        classes: Sequence[int] | dict[int, str] | None = None,
        classes_kwargs: dict[str, Any] = {},
        is_prob: bool = False,
        prob_labels: list[str] | None = None,
        prob_colors: list[ColorLike] | None = None,
        zorder: int | float | None = None,
        label_length: int = 20,
        **kwargs,
    ) -> "LineFigure":
        _zorder: float = 2.0
        if isinstance(zorder, (int, float)):
            _zorder = float(zorder)
        # Parse colors
        color = Color.from_optional(color)
        selection_color = Color.from_optional(selection_color)
        selection_highlight_color = Color.from_optional(selection_highlight_color)

        if isinstance(mode, str):
            if mode in ["line", "scatter", "area"]:
                self.mode = mode
            else:
                raise ValueError(
                    f'invalid `mode` "{mode}", expected either "line", "scatter" or "area"'
                )

        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
        else:
            value_range = (None, None)

        if isinstance(norm, Normalize):
            if log_scale == True and not isinstance(norm, LogNorm):
                norm = LogNorm(norm.vmin, norm.vmax)
            elif log_scale == False and isinstance(norm, LogNorm):
                norm = Normalize(norm.vmin, norm.vmax)
            if value_range[0] is not None:
                norm.vmin = value_range[0]  # type: ignore # FIXME
            if value_range[1] is not None:
                norm.vmax = value_range[1]  # type: ignore # FIXME
        else:
            if log_scale == True:
                norm = LogNorm(value_range[0], value_range[1])  # type: ignore # FIXME
            else:
                norm = Normalize(value_range[0], value_range[1])  # type: ignore # FIXME
        value_range = (norm.vmin, norm.vmax)

        values = np.asarray(values)
        time = np.asarray(time)
        if latitude is not None:
            latitude = np.asarray(latitude)
        if longitude is not None:
            longitude = np.asarray(longitude)

        # Validate inputs
        if is_prob:
            if len(values.shape) != 2:
                raise ValueError(
                    f"Since {is_prob=} values must be 2D, but has shape={values.shape}"
                )
        elif len(values.shape) != 1:
            raise ValueError(
                f"Since {is_prob=} values must be 1D, but has shape={values.shape}"
            )

        tmin_original = np.datetime64(to_timestamp(time[0]))
        tmax_original = np.datetime64(to_timestamp(time[-1]))
        vmin_original = values[0]
        vmax_original = values[-1]

        if selection_time_range is not None:
            self.selection_time_range = validate_time_range(selection_time_range)
            _selection_max_time_margin: tuple[pd.Timedelta, pd.Timedelta] | None = None
            if isinstance(selection_max_time_margin, (Sequence, np.ndarray)):
                _selection_max_time_margin = (
                    to_timedelta(selection_max_time_margin[0]),
                    to_timedelta(selection_max_time_margin[1]),
                )
            elif selection_max_time_margin is not None:
                _selection_max_time_margin = (
                    to_timedelta(selection_max_time_margin),
                    to_timedelta(selection_max_time_margin),
                )

            if _selection_max_time_margin is not None:
                time_range = [
                    np.max(
                        [
                            time[0],
                            (
                                self.selection_time_range[0]
                                - _selection_max_time_margin[0]
                            ).to_datetime64(),
                        ]
                    ),
                    np.min(
                        [
                            time[-1],
                            (
                                self.selection_time_range[1]
                                + _selection_max_time_margin[1]
                            ).to_datetime64(),
                        ]
                    ),
                ]

        if time_range is not None:
            if isinstance(time_range, Iterable) and len(time_range) == 2:
                for i in [0, -1]:
                    time_range = list(time_range)
                    if time_range[i] is None:
                        time_range[i] = time[i]
                    time_range = tuple(time_range)  # type: ignore
            time_range = (
                np.datetime64(to_timestamp(time_range[0])),
                np.datetime64(to_timestamp(time_range[-1])),
            )
            self.tmin = time_range[0]
            self.tmax = time_range[1]
        else:
            time_range = (time[0], time[-1])
        time_range = (
            np.datetime64(to_timestamp(time_range[0])),
            np.datetime64(to_timestamp(time_range[-1])),
        )

        _value_range: tuple[float, float] = select_value_range(
            data=values,
            value_range=value_range,
            pad_frac=0.0,
            use_min_max=True,
        )

        tmin = self.tmin or np.datetime64(time_range[0])
        tmax = self.tmax or np.datetime64(time_range[1])

        vmin: float = _value_range[0]
        vmax: float = _value_range[1]

        x: NDArray = time
        y: NDArray = values

        if is_prob:
            if label is None:
                label = "Probability"
            plot_stacked_propabilities(
                ax=self.ax,
                probabilities=values,
                time=time,
                labels=prob_labels,
                colors=prob_colors,
                zorder=_zorder,
                ax_label=label,
            )
            vmin = 0
            vmax = 1
        elif classes is not None:
            _yaxis_position = classes_kwargs.get("yaxis_position", "left")
            _is_left = _yaxis_position == "left"
            _label = format_var_label(label, units, label_len=label_length)

            plot_1d_integer_flag(
                ax=self.ax if _is_left else self.ax_right,
                ax2=self.ax_right if _is_left else self.ax,
                data=y,
                x=x,
                classes=classes,
                ax_label=_label,
                zorder=_zorder,
                **classes_kwargs,
            )
        else:
            line: list[Line2D] | PathCollection | PolyCollection
            if "line" in self.mode:
                line = self.ax.plot(
                    x,
                    y,
                    marker="none",
                    linewidth=linewidth,
                    linestyle=linestyle,
                    color=color,
                    alpha=alpha,
                    zorder=_zorder,
                )
            elif "scatter" in self.mode:
                line = self.ax.scatter(
                    x,
                    y,
                    marker=marker,
                    s=markersize,
                    color=color,
                    alpha=alpha,
                    zorder=_zorder,
                )
            elif "area" in self.mode:
                line = self.ax.fill_between(
                    x,
                    [0] * x.shape[0],
                    y,
                    color=color,
                    alpha=alpha,
                    zorder=zorder or 0.0,
                )
            else:
                raise ValueError(f"invalid `mode` {self.mode}")

            format_numeric_ticks(
                ax=self.ax,
                axis="y",
                label=format_var_label(label, units, label_len=label_length),
                max_line_length=label_length,
                show_label=self.show_value_left,
                show_values=self.show_value_left,
            )
            format_numeric_ticks(
                ax=self.ax_right,
                axis="y",
                label=format_var_label(label, units, label_len=label_length),
                max_line_length=label_length,
                show_label=self.show_value_right,
                show_values=self.show_value_right,
            )

        if selection_time_range is not None:
            if selection_highlight:
                if selection_highlight_inverted:
                    self.ax.axvspan(
                        tmin,  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                        zorder=3,
                    )
                    self.ax.axvspan(
                        self.selection_time_range[1],  # type: ignore
                        tmax,  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                        zorder=3,
                    )
                else:
                    self.ax.axvspan(
                        self.selection_time_range[0],  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                        zorder=3,
                    )

            for t in self.selection_time_range:  # type: ignore
                self.ax.axvline(
                    x=t,  # type: ignore
                    color=selection_color,
                    linestyle=selection_linestyle,
                    linewidth=selection_linewidth,
                    zorder=3,
                )

        self._set_axes(
            tmin=tmin,
            tmax=tmax,
            vmin=vmin,
            vmax=vmax,
            time=time,
            tmin_original=tmin_original,
            tmax_original=tmax_original,
            latitude=latitude,
            longitude=longitude,
            ax_style_top=ax_style_top,
            ax_style_bottom=ax_style_bottom,
        )

        if mark_profiles_at is not None:
            for t in to_timestamps(mark_profiles_at):
                self.ax.axvline(
                    t,  # type: ignore
                    color=selection_color,
                    linestyle="solid",
                    linewidth=selection_linewidth,
                    zorder=3,
                )

        return self

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str,
        *,
        time_var: str = TIME_VAR,
        lat_var: str = TRACK_LAT_VAR,
        lon_var: str = TRACK_LON_VAR,
        along_track_dim: str = ALONG_TRACK_DIM,
        values: NDArray | None = None,
        time: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        site: str | GroundSite | None = None,
        radius_km: float = 100.0,
        mark_closest_profile: bool = False,
        show_info: bool = True,
        info_text_loc: str | None = None,
        # Common args for wrappers
        mode: str | Literal["line", "scatter", "area"] | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        color: str | None = Color("ec:blue"),
        alpha: float = 1.0,
        linestyle: str | None = "solid",
        linewidth: float | int | None = 2.0,
        marker: str | None = "s",
        markersize: float | int | None = 2.0,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str | None = Color("white"),
        selection_highlight_alpha: float = 0.5,
        selection_max_time_margin: (
            TimedeltaLike | Sequence[TimedeltaLike] | None
        ) = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        mark_profiles_at: Sequence[TimestampLike] | None = None,
        classes: Sequence[int] | dict[int, str] | None = None,
        classes_kwargs: dict[str, Any] = {},
        is_prob: bool = False,
        prob_labels: list[str] | None = None,
        prob_colors: list[ColorLike] | None = None,
        zorder: int | float | None = None,
        label_length: int = 20,
        **kwargs,
    ) -> "LineFigure":
        # Collect all common args for wrapped plot function call
        local_args = locals()
        # Delete all args specific to this wrapper function
        del local_args["self"]
        del local_args["ds"]
        del local_args["var"]
        del local_args["time_var"]
        del local_args["lat_var"]
        del local_args["lon_var"]
        del local_args["along_track_dim"]
        del local_args["site"]
        del local_args["radius_km"]
        del local_args["show_info"]
        del local_args["info_text_loc"]
        del local_args["mark_closest_profile"]
        # Delete kwargs to then merge it with the residual common args
        del local_args["kwargs"]
        all_args = {**local_args, **kwargs}

        if all_args["values"] is None:
            all_args["values"] = ds[var].values
        if all_args["time"] is None:
            all_args["time"] = ds[time_var].values
        if all_args["latitude"] is None:
            all_args["latitude"] = ds[lat_var].values
        if all_args["longitude"] is None:
            all_args["longitude"] = ds[lon_var].values

        # Set default values depending on variable name
        if label is None:
            all_args["label"] = (
                "Values" if not hasattr(ds[var], "long_name") else ds[var].long_name
            )
        if units is None:
            all_args["units"] = "-" if not hasattr(ds[var], "units") else ds[var].units
        if classes is not None and len(classes) > 0:
            all_args["value_range"] = (-0.5, len(classes) - 0.5)
        elif value_range is None and log_scale is None and norm is None:
            all_args["norm"] = get_default_norm(var, file_type=ds)

        # Handle overpass
        _site: GroundSite | None = None
        if isinstance(site, GroundSite):
            _site = site
        elif isinstance(site, str):
            _site = get_ground_site(site)
        else:
            pass

        if isinstance(_site, GroundSite):
            info_overpass = get_overpass_info(
                ds,
                radius_km=radius_km,
                site=_site,
                time_var=time_var,
                lat_var=lat_var,
                lon_var=lon_var,
                along_track_dim=along_track_dim,
            )
            overpass_time_range = info_overpass.time_range
            all_args["selection_time_range"] = overpass_time_range
            if mark_closest_profile:
                _mark_profiles_at = all_args["mark_profiles_at"]
                if isinstance(_mark_profiles_at, (Sequence, np.ndarray)):
                    list(_mark_profiles_at).append(info_overpass.closest_time)
                    all_args["mark_profiles_at"] = _mark_profiles_at
                else:
                    all_args["mark_profiles_at"] = [info_overpass.closest_time]

        self.plot(**all_args)

        self._set_info_text_loc(info_text_loc)
        if show_info:
            self.info_text = add_text_product_info(
                self.ax, ds, append_to=self.info_text, loc=self.info_text_loc
            )

        return self

    def to_texture(self) -> "LineFigure":
        # Remove anchored text and other artist text objects
        for artist in reversed(self.ax.artists):
            if isinstance(artist, (Text, AnchoredOffsetbox)):
                artist.remove()

        # Completely remove axis ticks and labels
        self.ax.axis("off")

        if self.ax_top:
            self.ax_top.axis("off")

        if self.ax_right:
            self.ax_right.axis("off")

        # Remove white frame around figure
        self.fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        return self

    def invert_xaxis(self) -> "LineFigure":
        """Invert the x-axis."""
        self.ax.invert_xaxis()
        if self.ax_top:
            self.ax_top.invert_xaxis()
        return self

    def invert_yaxis(self) -> "LineFigure":
        """Invert the y-axis."""
        self.ax.invert_yaxis()
        if self.ax_right:
            self.ax_right.invert_yaxis()
        return self

    def show(self) -> None:
        import IPython
        import matplotlib.pyplot as plt
        from IPython.display import display

        if IPython.get_ipython() is not None:
            display(self.fig)
        else:
            plt.show()

    def save(
        self,
        filename: str = "",
        filepath: str | None = None,
        ds: xr.Dataset | None = None,
        ds_filepath: str | None = None,
        dpi: float | Literal["figure"] = "figure",
        orbit_and_frame: str | None = None,
        utc_timestamp: TimestampLike | None = None,
        use_utc_creation_timestamp: bool = False,
        site_name: str | None = None,
        hmax: int | float | None = None,
        radius: int | float | None = None,
        extra: str | None = None,
        transparent_outside: bool = False,
        verbose: bool = True,
        print_prefix: str = "",
        create_dirs: bool = False,
        transparent_background: bool = False,
        resolution: str | None = None,
        **kwargs,
    ) -> None:
        """
        Save a figure as an image or vector graphic to a file and optionally format the file name in a structured way using EarthCARE metadata.

        Args:
            figure (Figure | HasFigure): A figure object (`matplotlib.figure.Figure`) or objects exposing a `.fig` attribute containing a figure (e.g., `CurtainFigure`).
            filename (str, optional): The base name of the file. Can be extended based on other metadata provided. Defaults to empty string.
            filepath (str | None, optional): The path where the image is saved. Can be extended based on other metadata provided. Defaults to None.
            ds (xr.Dataset | None, optional): A EarthCARE dataset from which metadata will be taken. Defaults to None.
            ds_filepath (str | None, optional): A path to a EarthCARE product from which metadata will be taken. Defaults to None.
            pad (float, optional): Extra padding (i.e., empty space) around the image in inches. Defaults to 0.1.
            dpi (float | 'figure', optional): The resolution in dots per inch. If 'figure', use the figure's dpi value. Defaults to None.
            orbit_and_frame (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            utc_timestamp (TimestampLike | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            use_utc_creation_timestamp (bool, optional): Whether the time of image creation should be included in the file name. Defaults to False.
            site_name (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            hmax (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            radius (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            resolution (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            extra (str | None, optional): A custom string to be included in the file name. Defaults to None.
            transparent_outside (bool, optional): Whether the area outside figures should be transparent. Defaults to False.
            verbose (bool, optional): Whether the progress of image creation should be printed to the console. Defaults to True.
            print_prefix (str, optional): A prefix string to all console messages. Defaults to "".
            create_dirs (bool, optional): Whether images should be saved in a folder structure based on provided metadata. Defaults to False.
            transparent_background (bool, optional): Whether the background inside and outside of figures should be transparent. Defaults to False.
            **kwargs (dict[str, Any]): Keyword arguments passed to wrapped function call of `matplotlib.pyplot.savefig`.
        """
        save_plot(
            fig=self.fig,
            filename=filename,
            filepath=filepath,
            ds=ds,
            ds_filepath=ds_filepath,
            dpi=dpi,
            orbit_and_frame=orbit_and_frame,
            utc_timestamp=utc_timestamp,
            use_utc_creation_timestamp=use_utc_creation_timestamp,
            site_name=site_name,
            hmax=hmax,
            radius=radius,
            extra=extra,
            transparent_outside=transparent_outside,
            verbose=verbose,
            print_prefix=print_prefix,
            create_dirs=create_dirs,
            transparent_background=transparent_background,
            resolution=resolution,
            **kwargs,
        )



import warnings
from typing import Callable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray
from xarray import Dataset

from ...constants import ALONG_TRACK_DIM, TRACK_LON_VAR
from ...xarray_utils import concat_datasets
from ._generic import read_product
from .file_info import ProductDataFrame


def circular_mean_np(data: NDArray, axis: int = -1, degrees: bool = True) -> NDArray:
    if degrees:
        data = np.deg2rad(data)
    sin_mean = np.mean(np.sin(data), axis=axis)
    cos_mean = np.mean(np.cos(data), axis=axis)
    angle = np.arctan2(sin_mean, cos_mean)
    return np.rad2deg(angle) if degrees else angle


def read_products(
    filepaths: Sequence[str] | NDArray[np.str_] | pd.DataFrame,
    zoom_at: float | None = None,
    along_track_dim: str = ALONG_TRACK_DIM,
    func: Callable | None = None,
    func_inputs: Sequence[dict] | None = None,
    max_num_files: int = 8,
    coarsen: bool = True,
) -> Dataset:
    """Read and concatenate a sequence of EarthCARE frames into a single xarray Dataset.

    By default, the dataset is coarsened according to the number of input frames (e.g.,
    combining 3 products averages every 3 profiles, so the along-track dimension remains
    comparable to a single product). Optionally applies a processing function to each
    frame and zooms in on a specific region (defined by `zoom_at`) without coarsening.
    Coarsening can also be turned of but might case memory issues.

    Args:
        filepaths (Sequence[str] or pandas.DataFrame):
            EarthCARE product file paths as a list or a DataFrame with metadata
            including `filepath`, `orbit_number`, and `frame_id`.
        zoom_at (float, optional):
            If set, selects only a zoomed-in portion of the frames around this
            fractional index. Defaults to None.
        along_track_dim (str, optional):
            Name of the dimension to concatenate along. Defaults to ALONG_TRACK_DIM.
        func (Callable, optional):
            Function to apply to each frame after loading. Defaults to None.
        func_inputs (Sequence[dict], optional):
            Optional per-frame arguments to pass to `func`. Defaults to None.
        max_num_files (int, optional):
            Max. number of files that are allowed to be loaded at once.
            A `ValueError` is raised if above. Defaults to 8 (e.g., full orbit).
        coarsen (bool, optional):
            If Ture, read data sets are coarened depending on the number given of files.
            Only aplicable when not zooming. Defaults to Ture.

    Returns:
        Dataset: Concatenated dataset with all frames along `along_track_dim`.
    """
    if isinstance(filepaths, str):
        filepaths = [filepaths]
    elif isinstance(filepaths, pd.DataFrame):
        df = filepaths.sort_values(by="orbit_and_frame")
        filepaths = df["filepath"].tolist()
    else:
        df = ProductDataFrame.from_files(list(filepaths)).sort_values(
            by="orbit_and_frame"
        )
        df.validate_columns()
        filepaths = df["filepath"].tolist()

    if len(filepaths) == 0:
        raise ValueError(f"Given sequence of product files paths is empty")
    elif len(filepaths) == 1:
        warnings.warn(f"Can not concatenate frames since only one file path was given")
        return read_product(filepaths[0])
    elif len(filepaths) > max_num_files:
        raise ValueError(
            f"Too many files provided: {len(filepaths)} (currently maximum allowed is {max_num_files}). "
            "Please reduce the number of files or increase the allowed amount by setting the argument max_num_files."
        )
    elif len(filepaths) > 8:
        warnings.warn(
            f"You provided {len(filepaths)} files, which is more than one full orbit (8 files). "
            "Processing might take longer than usual."
        )

    # # Construct filename suffix from orbit/frame numbers
    # orbit_start = str(df["orbit_number"].iloc[0]).zfill(5)
    # orbit_end = str(df["orbit_number"].iloc[-1]).zfill(5)
    # frame_start = df["frame_id"].iloc[0]
    # frame_end = df["frame_id"].iloc[-1]

    # if orbit_start == orbit_end:
    #     oaf_string = (
    #         f"{orbit_start}{frame_start}"
    #         if frame_start == frame_end
    #         else f"{orbit_start}{frame_start}-{frame_end}"
    #     )
    # else:
    #     oaf_string = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"

    def apply_func(ds: Dataset, i: int) -> Dataset:
        """Apply a processing function to a dataset if specified."""
        if func is None:
            return ds
        if func_inputs is None:
            return func(ds)
        if i < len(func_inputs):
            return func(ds, **func_inputs[i])
        raise IndexError("Too few function inputs provided")

    num_files = len(filepaths)
    ds: xr.Dataset | None = None

    if zoom_at is not None:
        # Zoomed read: select portions of two adjacent frames
        frame_indices = np.unique([int(np.floor(zoom_at)), int(np.ceil(zoom_at))])
        offset = zoom_at - frame_indices[0]
        filepaths = [filepaths[i] for i in frame_indices]

        for i, filepath in enumerate(filepaths):
            with read_product(filepath) as frame_ds:
                frame_ds = apply_func(frame_ds, frame_indices[i])

                # Preserve original dtypes
                original_dtypes = {v: frame_ds[v].dtype for v in frame_ds.variables}

                # Select relevant portion of the frame
                n = len(frame_ds[along_track_dim])
                sel_slice = (
                    slice(int(np.floor(n * offset)), n)
                    if i == 0
                    else slice(0, int(np.ceil(n * offset)))
                )
                frame_ds = frame_ds.sel({along_track_dim: sel_slice})

                # Restore dtypes
                for v, dtype in original_dtypes.items():
                    frame_ds[v] = frame_ds[v].astype(dtype)

                ds = (
                    frame_ds.copy()
                    if ds is None
                    else concat_datasets(
                        ds.copy(), frame_ds.copy(), dim=along_track_dim
                    )
                )

    else:
        # Full read and coarsen each frame
        for i, filepath in enumerate(filepaths):
            with read_product(filepath) as frame_ds:
                frame_ds = apply_func(frame_ds, i)

                if coarsen:
                    original_dtypes = {v: frame_ds[v].dtype for v in frame_ds.variables}

                    coarsen_dims = {along_track_dim: num_files}

                    # Circular mean for longitude
                    lon_coarse = (
                        frame_ds["longitude"]
                        .coarsen(coarsen_dims, boundary="trim")
                        .reduce(circular_mean_np)
                    )
                    _tmp_attrs = lon_coarse.attrs
                    lon_coarse.attrs = {}

                    # Regular mean for the rest
                    rest = (
                        frame_ds.drop_vars("longitude")
                        .coarsen(coarsen_dims, boundary="trim")
                        .mean()  # type: ignore
                    )

                    # Merge results
                    frame_ds = xr.merge([lon_coarse, rest])
                    frame_ds["longitude"].attrs = _tmp_attrs

                    for v, dtype in original_dtypes.items():
                        frame_ds[v] = frame_ds[v].astype(dtype)

                ds = (
                    frame_ds
                    if ds is None
                    else concat_datasets(ds, frame_ds, dim=along_track_dim)
                )

    # Set output file sources
    if isinstance(ds, Dataset):
        ds.encoding["sources"] = list(filepaths)
        return ds
    else:
        raise RuntimeError(f"Bad implementation")



from xarray import Dataset

from ...constants import *


def _rename(
    ds: Dataset,
    name: str | None,
    new_name: str,
) -> Dataset:
    if isinstance(name, str):
        ds = ds.rename({name: new_name})
    return ds


def rename_and_create_temperature_vars(
    ds: Dataset,
    temperature_var: str | None = None,
    is_kelvin: bool = True,
    var_name_kelvin: str = TEMP_KELVIN_VAR,
    var_name_celsius: str = TEMP_CELSIUS_VAR,
) -> Dataset:
    if isinstance(temperature_var, str):
        if is_kelvin:
            name1 = var_name_kelvin
            units1 = "K"
            name2 = var_name_celsius
            units2 = r"$^{\circ}$C"
            conversion = -273.15
        else:
            name1 = var_name_celsius
            units1 = r"$^{\circ}$C"
            name2 = var_name_kelvin
            units2 = "K"
            conversion = 273.15
        ds = _rename(ds, temperature_var, name1)
        ds = rename_var_info(ds, name1, "Temperature", "Temperature", units1)
        ds[name2] = ds[name1].copy() + conversion
        ds = rename_var_info(ds, name2, "Temperature", "Temperature", units2)
    return ds


def rename_common_dims_and_vars(
    ds: Dataset,
    along_track_dim: str | None = None,
    across_track_dim: str | None = None,
    vertical_dim: str | None = None,
    time_var: str | None = None,
    height_var: str | None = None,
    track_lat_var: str | None = None,
    track_lon_var: str | None = None,
    swath_lat_var: str | None = None,
    swath_lon_var: str | None = None,
    elevation_var: str | None = None,
    tropopause_var: str | None = None,
    temperature_var: str | None = None,
    land_flag_var: str | None = None,
) -> Dataset:
    """Renames standard dimensions and variables to create consistency across EarthCARE products."""
    ds = _rename(ds, along_track_dim, ALONG_TRACK_DIM)
    ds = _rename(ds, across_track_dim, ACROSS_TRACK_DIM)
    ds = _rename(ds, vertical_dim, VERTICAL_DIM)
    ds = _rename(ds, time_var, TIME_VAR)
    ds = _rename(ds, height_var, HEIGHT_VAR)
    ds = _rename(ds, track_lat_var, TRACK_LAT_VAR)
    ds = _rename(ds, track_lon_var, TRACK_LON_VAR)
    ds = _rename(ds, swath_lat_var, SWATH_LAT_VAR)
    ds = _rename(ds, swath_lon_var, SWATH_LON_VAR)
    ds = _rename(ds, elevation_var, ELEVATION_VAR)
    ds = _rename(ds, land_flag_var, LAND_FLAG_VAR)
    ds = _rename(ds, tropopause_var, TROPOPAUSE_VAR)
    ds = rename_and_create_temperature_vars(ds, temperature_var, is_kelvin=True)
    return ds


def rename_var_info(
    ds: Dataset,
    var: str,
    name: str | None = None,
    long_name: str | None = None,
    units: str | None = None,
) -> Dataset:
    if name is not None:
        ds[var] = ds[var].assign_attrs(name=name)
    if long_name is not None:
        if hasattr(ds[var], "long_name") and ds[var].long_name != long_name:
            ds[var] = ds[var].assign_attrs(original_long_name=ds[var].long_name)
        ds[var] = ds[var].assign_attrs(long_name=long_name)
    if units is not None:
        ds[var] = ds[var].assign_attrs(units=units)
    return ds



import re
from dataclasses import dataclass

import numpy as np
import pandas as pd
import xarray as xr


@dataclass
class _ParsedColumnName:
    old_name: str
    name: str
    long_name: str
    units: str


def _parse_units(units: str) -> str:

    pattern = re.compile(
        r"(?i)^((mm|msr)(\^?-1)\s*(m|sr)(\^?-1)|(m|sr)(\^?-1)\s*(mm|msr)(\^?-1))$"
    )
    match = pattern.match(units)
    if match:
        return "Mm-1 sr-1"

    pattern = re.compile(r"(?i)^((m)(\^?-1)\s*(sr)(\^?-1)|(sr)(\^?-1)\s*(m)(\^?-1))$")
    match = pattern.match(units)
    if match:
        return "m-1 sr-1"

    pattern = re.compile(r"(?i)^mm(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "Mm-1"

    pattern = re.compile(r"(?i)^m(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "m-1"

    pattern = re.compile(r"(?i)^msr(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "Msr-1"

    pattern = re.compile(r"(?i)^sr(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "sr-1"

    return units


def _parse_column_name(column_name: str) -> _ParsedColumnName:
    old_name: str = column_name
    name: str
    long_name: str
    units: str

    def _clean_string(s: str) -> str:
        s = re.sub(r"[^A-Za-z0-9]", "_", s.strip().lower())
        return re.sub(r"_+", "_", s)

    pattern = re.compile(r"^(.*?)\s*(?:\((.*?)\))?\:?$")
    match = pattern.match(column_name)
    if match:
        name = _clean_string(match.group(1))
        long_name = match.group(1).strip()
        units = match.group(2) if match.group(2) else ""
    else:
        name = _clean_string(column_name)
        long_name = column_name
        units = ""

    return _ParsedColumnName(
        old_name=old_name,
        name=name,
        long_name=long_name,
        units=_parse_units(units),
    )


def _make_column_names_unique(columns: list[str]) -> list[str]:
    seen: dict[str, int] = {}
    unique_columns: list[str] = []
    for col in columns:
        if col not in seen:
            seen[col] = 1
            unique_columns.append(col)
        else:
            count = seen[col]
            new_col = f"{col}_{count}"
            while new_col in seen:
                count += 1
                new_col = f"{col}_{count}"
            seen[col] = count + 1
            seen[new_col] = 1
            unique_columns.append(new_col)
    return unique_columns


def read_polly(input: str | xr.Dataset) -> xr.Dataset:
    """Reads manually processed PollyXT output text files as `xarray.Dataset` or returns an already open one."""

    if isinstance(input, xr.Dataset):
        return input

    with open(input, "r", encoding="utf-8", errors="ignore") as f:
        df = pd.read_csv(f, sep="\t")

    new_columns = [_parse_column_name(c) for c in df.columns]
    new_column_names = [c.name for c in new_columns]
    new_column_names = _make_column_names_unique(new_column_names)
    df.columns = pd.Index(new_column_names)

    ds = xr.Dataset.from_dataframe(df)
    ds = ds.assign_coords(index=ds.height.values)
    ds = ds.rename({"index": "vertical"})
    if "time" not in ds:
        ds = ds.assign({"time": np.datetime64("1970-01-01T00:00:00.000", "ms")})

    vars_order = ["time"] + [v for v in ds.data_vars if v != "time"]
    ds = ds[vars_order]

    for c in new_columns:
        if c.units == "km":
            ds[c.name].values = ds[c.name].values * 1e3
            c.units = c.units.replace("k", "")
        elif c.units in ["Mm-1 sr-1", "Mm-1", "Msr-1"]:
            ds[c.name].values = ds[c.name].values / 1e6
            c.units = c.units.replace("M", "")

        ds[c.name] = ds[c.name].assign_attrs(
            dict(
                long_name=c.long_name,
                units=c.units,
            )
        )
    return ds



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    BSC_LABEL,
    DEPOL_LABEL,
    EXT_LABEL,
    LR_LABEL,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..science_group import read_science_data


def _convert_depol(ds: xr.Dataset, var: str) -> xr.Dataset:
    ds[var].data = ds[var].data / 100.0
    ds[var] = ds[var].assign_attrs(
        {
            "units": "-",
            "valid_range": "[0, 1]",
            "earthcarekit": "Modified by earthcarekit: converted from percantage to decimal",
        }
    )
    return ds


def read_product_acla(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens ATL_CLA_2A file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.JAXA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    # Combine same dimensions
    ds = ds.squeeze()  # Removes dimensions of size 1
    ds = ds.rename({"phony_dim_7": "phony_dim_3"})
    ds = ds.rename({"phony_dim_9": "phony_dim_6"})
    ds = ds.rename({"phony_dim_8": "phony_dim_5"})

    # Rename content
    ds = ds.rename({"phony_dim_6": "along_track_2"})
    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="phony_dim_3",
        vertical_dim="phony_dim_5",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        elevation_var="surface_elevation",
        land_flag_var="land_water_flag",
    )

    ds = _convert_depol(ds, "aerosol_depolarization_10km")
    ds = _convert_depol(ds, "cloud_depolarization_1km")
    ds = _convert_depol(ds, "cloud_depolarization_10km")

    return ds



import numpy as np
import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
    UNITS_KELVIN,
    UNITS_MSI_RADIANCE,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    drop_samples_with_missing_geo_data_along_track,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def _get_rgb_from_swir1_nir_vis(
    ds: xr.Dataset,
    swir1_var: str = "swir1",
    nir_var: str = "nir",
    vis_var: str = "vis",
) -> np.ndarray:
    get_min_max = lambda x: np.array([ds[x].quantile(0.01), ds[x].quantile(0.99)])

    r_min, r_max = get_min_max(swir1_var)
    g_min, g_max = get_min_max(nir_var)
    b_min, b_max = get_min_max(vis_var)

    r_w, g_w, b_w = [1.0, 1.0, 1.0]
    r_s, g_s, b_s = [1.0, 1.0, 1.0]

    get_v = lambda x, _w, _s, _min, _max: np.clip(
        _w * (ds[x] - _min) / (_s * (_max - _min)), a_min=0, a_max=1
    ).T

    r_v = get_v(swir1_var, r_w, r_s, r_min, r_max)
    g_v = get_v(nir_var, g_w, g_s, g_min, g_max)
    b_v = get_v(vis_var, b_w, b_s, b_min, b_max)

    rgb = np.stack((r_v, g_v, b_v), axis=2)
    rgb[np.isnan(rgb)] = 0.0

    return rgb


def _add_rgb(
    ds: xr.Dataset,
    swir1_var: str = "swir1",
    nir_var: str = "nir",
    vis_var: str = "vis",
    rgb_var: str = "rgb",
    rgb_dim: str = "rgb_channel",
    along_track_dim: str = "along_track",
    across_track_dim: str = "across_track",
) -> xr.Dataset:
    rgb = _get_rgb_from_swir1_nir_vis(
        ds,
        swir1_var=swir1_var,
        nir_var=nir_var,
        vis_var=vis_var,
    )

    ds[rgb_var] = ((across_track_dim, along_track_dim, rgb_dim), rgb)
    ds[rgb_var].attrs["units"] = ""
    ds[rgb_var].attrs["long_name"] = "False RGB image"

    return ds


def _get_vns_name(
    wavelength: str, band_name: str | None = None, is_uncertainty: bool = False
) -> str:
    sub_str = "" if not is_uncertainty else "uncertainty "
    name: str = f"Radiance {sub_str}at {wavelength} nm"
    # if band_name:
    #     return name + f" ({band_name})"
    return name


def _get_tir_name(
    wavelength: str, band_name: str | None = None, is_uncertainty: bool = False
) -> str:
    sub_str = "" if not is_uncertainty else "uncertainty "
    name: str = f"BT {sub_str}at {wavelength} µm"
    # if band_name:
    #     return name + f" ({band_name})"
    return name


def read_product_mrgr(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens MSI_RGR_1C file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = drop_samples_with_missing_geo_data_along_track(
        ds=ds,
        swath_lat_var="latitude",
        along_track_dim="along_track",
        across_track_dim="across_track",
    )

    ds = ds.assign(
        vis=ds["pixel_values"].isel({"band": 0}),
        nir=ds["pixel_values"].isel({"band": 1}),
        swir1=ds["pixel_values"].isel({"band": 2}),
        swir2=ds["pixel_values"].isel({"band": 3}),
        tir1=ds["pixel_values"].isel({"band": 4}),
        tir2=ds["pixel_values"].isel({"band": 5}),
        tir3=ds["pixel_values"].isel({"band": 6}),
        vis_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 0}),
        nir_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 1}),
        swir1_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 2}),
        swir2_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 3}),
        tir1_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 4}),
        tir2_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 5}),
        tir3_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 6}),
        vis_line_quality_status=ds["line_quality_status"].isel({"band": 0}),
        nir_line_quality_status=ds["line_quality_status"].isel({"band": 1}),
        swir1_line_quality_status=ds["line_quality_status"].isel({"band": 2}),
        swir2_line_quality_status=ds["line_quality_status"].isel({"band": 3}),
        tir1_line_quality_status=ds["line_quality_status"].isel({"band": 4}),
        tir2_line_quality_status=ds["line_quality_status"].isel({"band": 5}),
        tir3_line_quality_status=ds["line_quality_status"].isel({"band": 6}),
    )

    for v in ["vis", "vis_uncertainty"]:
        _wavelength = "670"
        _name = "VIS"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_vns_name(_wavelength, None, _is_uncertainty),
            long_name=_get_vns_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_MSI_RADIANCE,
        )
    for v in ["nir", "nir_uncertainty"]:
        _wavelength = "865"
        _name = "NIR"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_vns_name(_wavelength, None, _is_uncertainty),
            long_name=_get_vns_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_MSI_RADIANCE,
        )
    for v in ["swir1", "swir1_uncertainty"]:
        _wavelength = "1650"
        _name = "SWIR-1"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_vns_name(_wavelength, None, _is_uncertainty),
            long_name=_get_vns_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_MSI_RADIANCE,
        )
    for v in ["swir2", "swir2_uncertainty"]:
        _wavelength = "2210"
        _name = "SWIR-2"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_vns_name(_wavelength, None, _is_uncertainty),
            long_name=_get_vns_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_MSI_RADIANCE,
        )
    for v in ["tir1", "tir1_uncertainty"]:
        _wavelength = "8.8"
        _name = "TIR-1"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_tir_name(_wavelength, None, _is_uncertainty),
            long_name=_get_tir_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_KELVIN,
        )
    for v in ["tir2", "tir2_uncertainty"]:
        _wavelength = "10.8"
        _name = "TIR-2"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_tir_name(_wavelength, None, _is_uncertainty),
            long_name=_get_tir_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_KELVIN,
        )
    for v in ["tir3", "tir3_uncertainty"]:
        _wavelength = "12.0"
        _name = "TIR-3"
        _is_uncertainty = "uncertainty" in v
        ds = rename_var_info(
            ds=ds,
            var=v,
            name=_get_tir_name(_wavelength, None, _is_uncertainty),
            long_name=_get_tir_name(_wavelength, _name, _is_uncertainty),
            units=UNITS_KELVIN,
        )

    ds = ds.drop_vars(
        ["pixel_values", "pixel_values_uncertainty", "line_quality_status"]
    )
    ds = ds.drop_dims("band")

    ds = _add_rgb(ds)

    nadir_idx = get_nadir_index(ds, nadir_idx=270)

    ds = ds.rename({"latitude": SWATH_LAT_VAR})
    ds = ds.rename({"longitude": SWATH_LON_VAR})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var=SWATH_LAT_VAR, swath_lon_var=SWATH_LON_VAR
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        time_var="time",
    )

    ds = ds.reset_coords()

    return ds



from ._plot_line_between_figures import plot_line_between_figures
from .curtain import CurtainFigure
from .figure_type import ECKFigure, FigureType
from .line import LineFigure
from .map import MapFigure
from .multi_panel import create_column_figure_layout, create_multi_figure_layout
from .profile import ProfileFigure
from .swath import SwathFigure



import numpy as np
import xarray as xr


def _convert_all_fill_values_to_nan(ds: xr.Dataset) -> xr.Dataset:
    for v in ds.variables:
        if not hasattr(ds[v].encoding, "_FillValue"):
            if np.issubdtype(ds[v].values.dtype, np.floating):
                ds[v].values[ds[v].values >= 9.969209968386869e36] = np.nan
        if not hasattr(ds[v].encoding, "dtype"):
            _dtype = ds[v].encoding["dtype"]
            if np.issubdtype(_dtype, np.integer):
                if hasattr(ds[v].encoding, "_FillValue"):
                    fill_value = int(ds[v].encoding["_FillValue"])
                else:
                    fill_value = np.iinfo(_dtype).min
                ds[v].values[np.isnan(ds[v].values)] = fill_value
                ds[v] = ds[v].astype(_dtype)
    return ds



import re

import matplotlib.ticker as ticker
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from numpy.typing import NDArray

from ...utils.math_utils import get_decimal_shift
from ...utils.np_array_utils import (
    get_number_range,
    ismonotonic,
    isndarray,
    lookup_value_by_number,
)
from ...utils.time import (
    TimestampLike,
    get_time_range,
    lookup_value_by_timestamp,
    to_timestamps,
)
from ...utils.validation import validate_completeness_of_args
from .annotation import format_var_label
from .axis import AxisInput, validate_axis_input


def get_nice_ticks(
    values: NDArray,
    max_ticks: int = 10,
    ax_values: NDArray | None = None,
):
    """
    Finds nice tick values based on monotonically increasing data.

    Parameters:
        values (numpy.ndarray): Monotonically increasing data array (can be non-uniform).
        max_ticks (int, optional): Maximum number of ticks to return. Defaults to 10.
        ax_values (numpy.ndarray | None, optional): Optional array of same shape as `values`
            containing data of an axis the `values` will be displayed on. If this parameter
            is given, the 'nice' or rounded label from `values` will be displayed at the
            closest index of `ax_values`.

    Returns:
        ticks (tuple[numpy.ndarray, numpy.ndarray]):
            - tick_positions: An array containing the ticks locations.
            - ticks_labels: An array containing the respective tick labels.
    """
    values = np.asarray(values)
    if ax_values is not None:
        ax_values = np.asarray(ax_values)
    ismonotonic(values, raise_error=True)

    min_val = values[0]
    max_val = values[-1]

    is_increasing = min_val < max_val
    if not is_increasing:
        values = values[::-1]
        min_val = values[0]
        max_val = values[-1]
        if isinstance(ax_values, np.ndarray):
            ax_values = ax_values[::-1]

    raw_step_size = (max_val - min_val) / max_ticks
    decimal_shift = get_decimal_shift(raw_step_size)

    values = np.array(values) * 10**decimal_shift
    min_val, max_val = values[0], values[-1]
    raw_step_size = (max_val - min_val) / max_ticks

    magnitude = 10 ** (len(str(int(np.floor(raw_step_size)))) - 1)

    for factor in [1, 2, 3, 4, 5, 10]:
        clean_step_size = factor * magnitude
        if (max_val - min_val) // clean_step_size <= max_ticks:
            break

    first_clean_tick = (
        ((min_val // clean_step_size) + 1) * clean_step_size
        if min_val % clean_step_size != 0
        else min_val
    )
    ticks = np.arange(first_clean_tick, max_val + clean_step_size, clean_step_size)
    ticks = ticks[ticks <= max_val]

    tick_positions = ticks
    if isndarray(ax_values):
        ax_values = np.array(ax_values)
        if values.shape != ax_values.shape:
            raise ValueError(
                f"Shape mismatch: the shape of 'values' is not {values.shape} is not the same as of 'ax_values' {ax_values.shape}"
            )
        tick_positions = np.array(
            [ax_values[np.searchsorted(values, tick)] for tick in ticks]
        )
    else:
        tick_positions = ticks * 10 ** (-decimal_shift)
    ticks_labels = np.array([int(t) for t in ticks]) * 10 ** (-decimal_shift)

    return tick_positions, ticks_labels


def add_ticks(
    ax: Axes,
    ax_data: NDArray,
    tick_data: NDArray,
    axis: AxisInput = "x",
    format_function=None,
    major_tick_count=10,
    minor_tick_count=5,
    fontsize=None,
    title=None,
    show_tick_labels=True,
    in_linspace=True,
    **kwargs,
) -> None:
    """Adds ticks to a given axis."""
    if not isinstance(ax, Axes):
        raise TypeError(
            f"{add_ticks.__name__}() for `ax` expected type '{Axes.__name__}' but got '{type(ax).__name__}' instead"
        )

    axis = validate_axis_input(axis)

    ax_data = np.asarray(ax_data)
    tick_data = np.asarray(tick_data)
    if isinstance(tick_data[0], TimestampLike):
        tick_data = np.asarray(to_timestamps(tick_data))

    periods = major_tick_count
    minor_periods = minor_tick_count * (major_tick_count - 1) + 1

    if format_function is None:
        if isndarray(tick_data, dtype=np.datetime64):
            format_function = lambda t: t.strftime("%Y-%m-%d\n%H:%M:%S")
        else:
            format_function = lambda x: "${:.1f}$".format(x)
    if not in_linspace:
        ax_ticks, tick_labels = get_nice_ticks(
            tick_data,
            major_tick_count,
            ax_values=ax_data,
        )
        ax_minor_ticks: list | NDArray | pd.DatetimeIndex = []
    elif isndarray(ax_data, dtype=np.datetime64):
        ax_ticks = get_time_range(ax_data[0], ax_data[-1], periods=periods)
        ticks = np.array(
            [lookup_value_by_timestamp(t, ax_data, tick_data) for t in ax_ticks]
        )
        ax_minor_ticks = get_time_range(ax_data[0], ax_data[-1], periods=minor_periods)
        tick_labels = [format_function(tl) for tl in ticks]
    else:
        ax_ticks = get_number_range(ax_data[0], ax_data[-1], periods=periods)
        ticks = np.array(
            [lookup_value_by_number(t, ax_data, tick_data) for t in ax_ticks]
        )
        ax_minor_ticks = get_number_range(
            ax_data[0], ax_data[-1], periods=minor_periods
        )
        tick_labels = [format_function(tl) for tl in ticks]

    arg_idxs = validate_completeness_of_args(
        add_ticks.__name__, ["tick_data"], [], **kwargs
    )

    for idx in arg_idxs:
        _tick_data = kwargs[f"tick_data{idx}"]
        _format_function = lambda lon: "${:.1f}$".format(lon)
        if (
            f"format_function{idx}" in kwargs
            and kwargs[f"format_function{idx}"] is not None
        ):
            _format_function = kwargs[f"format_function{idx}"]

        if isndarray(ax_data, dtype=np.datetime64):
            ax_ticks = get_time_range(ax_data[0], ax_data[-1], periods=periods)
            _ticks = np.array(
                [lookup_value_by_timestamp(t, ax_data, _tick_data) for t in ax_ticks]
            )
        else:
            ax_ticks = get_number_range(ax_data[0], ax_data[-1], periods=periods)
            _ticks = np.array(
                [lookup_value_by_number(t, ax_data, _tick_data) for t in ax_ticks]
            )
        _tick_labels = [_format_function(tl) for tl in _ticks]
        tick_labels = [f"{l1}\n{l2}" for l1, l2 in zip(tick_labels, _tick_labels)]

    ax_set_ticks = {"x": ax.set_xticks, "y": ax.set_yticks}[axis]
    ax_set_ticks(
        ax_ticks,
        labels=tick_labels if show_tick_labels else ["" for x in tick_labels],
        fontsize=fontsize,
    )
    ax_set_ticks(ax_minor_ticks, minor=True)

    if title is not None:
        ax_set_label = {"x": ax.set_xlabel, "y": ax.set_ylabel}[axis]  # type: ignore
        ax_set_label(title)


def format_numeric_ticks(
    ax: Axes,
    axis: str = "x",
    label: str | None = None,
    max_line_length: int | None = None,
    show_label: bool = True,
    show_values: bool = True,
) -> None:
    """Format numeric tick labels on a matplotlib axis, using scientific notation if appropriate.

    This function sets up a smart tick locator and scalar formatter for either the x- or y-axis.
    It also appends the axis offset (e.g., *10^3) to the axis label instead of displaying it separately.
    Optionally, the label can be wrapped to a maximum line length.

    Args:
        ax (plt.Axes): The matplotlib Axes object.
        axis (str, optional): The axis to format, either "x" or "y". Defaults to "x".
        label (str | None, optional): The axis label to use. If None, the current label is used.
        max_line_length (int | None, optional): Maximum line length for wrapping the label. Defaults to None.

    Returns:
        None
    """
    _axis = {"y": ax.yaxis, "x": ax.xaxis}[axis]
    if label is None:
        label = _axis.get_label().get_text()

    class Labeloffset:
        def __init__(self, ax, label="", axis="y"):
            self.axis = _axis
            self.label = label
            if self.update(None) is not None:
                ax.callbacks.connect(axis + "lim_changed", self.update)
            ax.figure.canvas.draw()
            self.update(None)

        def update(self, lim):
            fmt = self.axis.get_major_formatter()
            s = fmt.get_offset()
            s = s.replace("\u2212", "-")  # Replace unicode minus with ASCII
            math_text = re.sub(
                r"\$\\times\\mathdefault\{10\^\{([^}]*)\}\}\\mathdefault\{\}\$",
                r"$\\times$10$^{\1}$",
                s,
            )
            self.axis.offsetText.set_visible(False)
            if len(math_text) > 0:
                self.axis.set_label_text(self.label + " " + math_text)
            else:
                self.axis.set_label_text(self.label)

    locator = ticker.MaxNLocator(nbins="auto", min_n_ticks=4, steps=[1, 2, 2.5, 5, 10])
    _axis.set_major_locator(locator)

    formatter = ticker.ScalarFormatter(useMathText=True)
    formatter.set_powerlimits((-3, 3))
    _axis.set_major_formatter(formatter)
    if show_label:
        label = format_var_label(label, label_len=max_line_length)
        lo = Labeloffset(ax, label=label, axis=axis)
    else:
        _axis.set_label_text(None)  # type: ignore

    if not show_values:
        _axis.set_ticklabels([])



import os
from typing import Literal, TypeAlias

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.figure import Figure

from ...utils.config import read_config
from ...utils.ground_sites import get_ground_site
from ...utils.read.product.file_info import ProductDataFrame, get_product_infos
from ...utils.time import TimestampLike, time_to_iso, to_timestamp
from ...utils.typing import HasFigure
from .save_figure_with_auto_margins import save_figure_with_auto_margins


def create_filepath(
    filename: str = "",
    filepath: str | None = None,
    ds: xr.Dataset | None = None,
    ds_filepath: str | None = None,
    orbit_and_frame: str | None = None,
    utc_timestamp: TimestampLike | None = None,
    use_utc_creation_timestamp: bool = False,
    site_name: str | None = None,
    hmax: int | float | None = None,
    radius: int | float | None = None,
    extra: str | None = None,
    create_dirs: bool = False,
    resolution: str | None = None,
) -> str:
    if not isinstance(filepath, str):
        config = read_config()
        filepath = os.path.join(config.path_to_images, filename)
    else:
        filepath = os.path.join(filepath, filename)

    df: ProductDataFrame | None = None
    if isinstance(ds, xr.Dataset):
        df = get_product_infos(ds)
    elif isinstance(ds_filepath, str):
        df = get_product_infos(ds_filepath)

    _file_type: str | None = None
    if df is not None:
        orbit_and_frame = df.orbit_and_frame[0]
        utc_timestamp = df.start_sensing_time[0]
        _file_type = df.file_type[0]

    if filepath:
        filename_components = []
        if orbit_and_frame is not None:
            filename_components.append(orbit_and_frame)

        if _file_type is not None:
            filename_components.append(_file_type)

        if utc_timestamp is not None:
            utc_timestamp = time_to_iso(utc_timestamp, format="%Y%m%dT%H%M%S")
            filename_components.append(utc_timestamp)

        if use_utc_creation_timestamp == True:
            creation_timestamp = time_to_iso(
                pd.Timestamp.now().utcnow(), format="%Y%m%dT%H%M%S"
            )
            filename_components.append(creation_timestamp)

        if site_name is not None:
            try:
                site_name = get_ground_site(site_name).name
            except ValueError as e:
                pass
            filename_components.append(f"site{site_name}")

        if radius is not None:
            radius_string = "rad" + str(int(np.round(radius))) + "m"
            filename_components.append(radius_string)

        if hmax is not None:
            hmax_string = "upto" + str(int(np.round(hmax / 1000))) + "km"
            filename_components.append(hmax_string)

        if resolution is not None:
            if resolution == "" or "high" in resolution.lower():
                filename_components.append("HiRes")
            elif "medium" in resolution.lower():
                filename_components.append("MedRes")
            elif "low" in resolution.lower():
                filename_components.append("LowRes")
            else:
                filename_components.append(resolution)

        if extra is not None:
            filename_components.append(extra)

        basename = os.path.basename(filepath)
        if len(basename) > 0 and basename[0] != ".":
            filename_components.append(basename)
        new_basename = "_".join(filename_components)

        dirname = os.path.dirname(filepath)

        if create_dirs and not os.path.exists(dirname):
            os.makedirs(dirname)

        new_filepath = os.path.join(dirname, new_basename)
        new_filepath = os.path.abspath(new_filepath)

        filepath = os.path.abspath(filepath)
        new_filepath = os.path.abspath(new_filepath)

        if os.path.isdir(new_filepath):
            new_filepath = os.path.join(new_filepath, ".png")
        return new_filepath
    else:
        raise ValueError("missing filepath inputs")


def save_plot(
    fig: Figure | HasFigure,
    filename: str = "",
    filepath: str | None = None,
    ds: xr.Dataset | None = None,
    ds_filepath: str | None = None,
    pad: float = 0.1,
    dpi: float | Literal["figure"] = "figure",
    orbit_and_frame: str | None = None,
    utc_timestamp: TimestampLike | None = None,
    use_utc_creation_timestamp: bool = False,
    site_name: str | None = None,
    hmax: int | float | None = None,
    radius: int | float | None = None,
    resolution: str | None = None,
    extra: str | None = None,
    transparent_outside: bool = False,
    verbose: bool = True,
    print_prefix: str = "",
    create_dirs: bool = False,
    transparent_background: bool = False,
    **kwargs,
) -> None:
    """
    Save a figure as an image or vector graphic to a file and optionally format the file name in a structured way using EarthCARE metadata.

    Args:
        figure (Figure | HasFigure): A figure object (`matplotlib.figure.Figure`) or objects exposing a `.fig` attribute containing a figure (e.g., `CurtainFigure`).
        filename (str, optional): The base name of the file. Can be extended based on other metadata provided. Defaults to empty string.
        filepath (str | None, optional): The path where the image is saved. Can be extended based on other metadata provided. Defaults to None.
        ds (xr.Dataset | None, optional): A EarthCARE dataset from which metadata will be taken. Defaults to None.
        ds_filepath (str | None, optional): A path to a EarthCARE product from which metadata will be taken. Defaults to None.
        pad (float, optional): Extra padding (i.e., empty space) around the image in inches. Defaults to 0.1.
        dpi (float | 'figure', optional): The resolution in dots per inch. If 'figure', use the figure's dpi value. Defaults to None.
        orbit_and_frame (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        utc_timestamp (TimestampLike | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        use_utc_creation_timestamp (bool, optional): Whether the time of image creation should be included in the file name. Defaults to False.
        site_name (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        hmax (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        radius (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        resolution (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
        extra (str | None, optional): A custom string to be included in the file name. Defaults to None.
        transparent_outside (bool, optional): Whether the area outside figures should be transparent. Defaults to False.
        verbose (bool, optional): Whether the progress of image creation should be printed to the console. Defaults to True.
        print_prefix (str, optional): A prefix string to all console messages. Defaults to "".
        create_dirs (bool, optional): Whether images should be saved in a folder structure based on provided metadata. Defaults to False.
        transparent_background (bool, optional): Whether the background inside and outside of figures should be transparent. Defaults to False.
        **kwargs (dict[str, Any]): Keyword arguments passed to wrapped function call of `matplotlib.pyplot.savefig`.
    """
    if not isinstance(fig, Figure):
        fig = fig.fig

    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    try:
        if transparent_background:
            transparent_outside = True

        new_filepath = create_filepath(
            filename,
            filepath,
            ds,
            ds_filepath,
            orbit_and_frame,
            utc_timestamp,
            use_utc_creation_timestamp,
            site_name,
            hmax,
            radius,
            extra,
            create_dirs,
            resolution,
        )

        if transparent_outside:
            fig.patch.set_alpha(0)
        if transparent_background:
            for ax in fig.get_axes():
                ax.patch.set_alpha(0)

        if verbose:
            print(f"{print_prefix}Saving plot ...", end="\r")
        save_figure_with_auto_margins(
            fig,
            new_filepath,
            pad=pad,
            dpi=dpi,
            **kwargs,
        )

        # Restore original settings
        if transparent_outside:
            fig.patch.set_alpha(1)
        if transparent_background:
            for ax in fig.get_axes():
                ax.patch.set_alpha(1)

        _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
        _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
        if verbose:
            print(f"{print_prefix}Plot saved (time taken {_dtime}): <{new_filepath}>")

        # raise ValueError(f"hi")
    except ValueError as e:
        if verbose:
            print(f"{print_prefix}Did not create plot since an error occured: {e}")



from logging import Logger
from typing import Literal, Sequence

import numpy as np
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import CM_AS_INCH, DEFAULT_PROFILE_SHOW_STEPS, TIME_VAR
from ....utils.time import TimedeltaLike, TimeRangeLike
from ....utils.typing import DistanceRangeLike
from ....utils.xarray_utils import filter_radius, filter_time
from ...figure import (
    CurtainFigure,
    ECKFigure,
    FigureType,
    MapFigure,
    ProfileFigure,
    create_multi_figure_layout,
)
from .._quicklook_results import QuicklookFigure
from ..set_default_height_range import set_none_height_range_to_default
from ._ecql_atl_ebd_2a import ecquicklook_aebd


def ecquicklook_aaer(
    ds: xr.Dataset,
    vars: list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = True,
    site: str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = (0, 30e3),
    ds_tropopause: xr.Dataset | None = None,
    ds_elevation: xr.Dataset | None = None,
    ds_temperature: xr.Dataset | None = None,
    resolution: Literal["low", "medium", "high", "l", "m", "h"] = "medium",
    closest_profile: bool = True,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
    mode: Literal["fast", "exact"] = "fast",
) -> QuicklookFigure:
    return ecquicklook_aebd(
        ds=ds,
        vars=vars,
        show_maps=show_maps,
        show_zoom=show_zoom,
        show_profile=show_profile,
        site=site,
        radius_km=radius_km,
        time_range=time_range,
        height_range=height_range,
        ds_tropopause=ds_tropopause,
        ds_elevation=ds_elevation,
        ds_temperature=ds_temperature,
        resolution="high",
        closest_profile=closest_profile,
        logger=logger,
        log_msg_prefix=log_msg_prefix,
        selection_max_time_margin=selection_max_time_margin,
        show_steps=show_steps,
        mode=mode,
    )



import datetime
from dataclasses import dataclass
from typing import Any, Iterable, Literal, Sequence, TypeAlias

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray

TimestampLike: TypeAlias = (
    str | np.str_ | pd.Timestamp | np.datetime64 | datetime.datetime
)

TimedeltaLike: TypeAlias = (
    str | np.str_ | pd.Timedelta | np.timedelta64 | datetime.timedelta
)

TimeRangeLike: TypeAlias = (
    tuple[TimestampLike, TimestampLike] | list[TimestampLike] | NDArray[np.datetime64]
)


def validate_time_range(
    time_range: TimeRangeLike | None,
) -> tuple[pd.Timestamp, pd.Timestamp]:
    """
    Checks if given `time_range` is valid, if True returns is as a `tuple`
    of two `pandas.Timestamp`s, if False raises `ValueError`.
    """
    if isinstance(time_range, (Iterable, np.ndarray)):
        if len(time_range) == 2:
            if all([isinstance(t, TimestampLike) for t in time_range]):
                result_time_range: tuple[pd.Timestamp, pd.Timestamp] = (
                    to_timestamp(time_range[0]),
                    to_timestamp(time_range[1]),
                )
                if result_time_range[0] <= result_time_range[1]:
                    return result_time_range
                else:
                    raise ValueError(
                        f"time range needs to increase monotonically, got instead: {tuple([str(t) for t in time_range])}"
                    )
    raise ValueError(f"invalid time range: {time_range}")


def time_to_string(time: TimestampLike, format: str = "%Y-%m-%dT%H:%M:%S") -> str:
    """Converts timestamp into a formatted string (defaults to YYYY-MM-DDTHH:MM:SS)."""
    timestamp = to_timestamp(time)
    if isinstance(timestamp, pd.Timestamp):
        return timestamp.strftime(format)
    else:
        return ""


def time_to_iso(
    time: TimestampLike,
    format: Literal["datetime", "date", "time"] | str = "datetime",
) -> str:
    """Converts timestamp into an ISO-formatted string.

    Args:
        time (TimestampLike): The input timestamp.
        format (Literal["date", "time", "datetime"]): The type of ISO string to return:
            - "datetime" -> YYYY-MM-DDTHH:MM:SS (default)
            - "date" -> YYYY-MM-DD
            - "time" -> HH:MM:SS

    Returns:
        str: ISO-formatted string corresponding to the requested type.
    """
    if not isinstance(format, str):
        raise TypeError(
            f"""Invalid type for output_format '{type(format).__name__}', expected Literal["date", "time", "datetime"]"""
        )
    if format == "date":
        return time_to_string(time, "%Y-%m-%d")
    elif format == "time":
        return time_to_string(time, "%H:%M:%S")
    elif format == "datetime":
        return time_to_string(time, "%Y-%m-%dT%H:%M:%S")
    else:
        return time_to_string(time, format)


def to_timestamp(t: TimestampLike, keep_tzinfo: bool = False) -> pd.Timestamp:
    """Converts input to `pandas.Timestamp`."""
    if isinstance(t, np.str_):
        t = str(t)

    if isinstance(t, TimestampLike):
        new_t = pd.Timestamp(t)
        if not keep_tzinfo and new_t.tzinfo is not None:
            new_t = new_t.tz_convert(None)
        return new_t
    else:
        raise TypeError(f"Input timestamp has invalud type ({type(t)}: {t})")


def to_timestamps(
    times: pd.DatetimeIndex | Sequence[TimestampLike] | ArrayLike,
    keep_tzinfo: bool = False,
) -> pd.DatetimeIndex:
    """Converts inputs to `pandas.Timestamp`s and returns them as `pandas.DatetimeIndex`."""
    if isinstance(times, pd.DatetimeIndex):
        return times
    if isinstance(times, (Sequence, np.ndarray)):
        return pd.DatetimeIndex(
            [to_timestamp(t, keep_tzinfo=keep_tzinfo) for t in times]
        )
    else:
        raise TypeError(f"Input timestamps has invalid type ({type(times)}: {times})")


def to_timedelta(t: TimedeltaLike) -> pd.Timedelta:
    """Converts input to `pandas.Timedelta`."""
    if isinstance(t, np.str_):
        t = str(t)

    if isinstance(t, TimedeltaLike):
        return pd.Timedelta(t)
    else:
        raise TypeError(f"Input timedelta has invalud type ({type(t)}: {t})")


def to_timedeltas(times: Sequence[TimedeltaLike] | NDArray) -> pd.TimedeltaIndex:
    """Converts inputs to `pandas.Timedelta`s and returns them as `pandas.TimedeltaIndex`."""
    if isinstance(times, (Sequence | np.ndarray)):
        return pd.TimedeltaIndex([to_timedelta(t) for t in times])
    else:
        raise TypeError(f"Input timestamps has invalud type ({type(times)}: {times})")


def format_time_range_text(
    st: TimestampLike, et: TimestampLike, show_date: bool = True, show_time: bool = True
) -> str:
    """Returns a formatted text label (`str`) describing the time range given by the start time `st` and end time `et`."""
    st = to_timestamp(st)
    et = to_timestamp(et)

    if (
        st.year == et.year
        and st.month == et.month
        and st.day == et.day
        and st.hour == et.hour
        and st.minute == et.minute
    ):
        texts = []
        if show_date:
            texts.append(f"{st.day} {st.strftime('%b')} {st.year}")
        if show_time:
            texts.append(f"{str(st.hour).zfill(2)}:{str(st.minute).zfill(2)} UTC")
        return ", ".join(texts)

    if (
        st.year == et.year
        and st.month == et.month
        and st.day == et.day
        and st.hour == et.hour
    ):
        texts = []
        if show_date:
            texts.append(f"{st.day} {st.strftime('%b')} {st.year}")
        if show_time:
            texts.append(
                f"{str(st.hour).zfill(2)}:{str(st.minute).zfill(2)}-{str(et.minute).zfill(2)} UTC"
            )
        return ", ".join(texts)

    if st.year == et.year and st.month == et.month and st.day == et.day:
        texts = []
        if show_date:
            texts.append(f"{st.day} {st.strftime('%b')} {st.year}")
        if show_time:
            texts.append(
                f"{str(st.hour).zfill(2)}:{str(st.minute).zfill(2)}-{str(et.hour).zfill(2)}:{str(et.minute).zfill(2)} UTC"
            )
        return ", ".join(texts)

    if st.year == et.year and st.month == et.month:
        if show_time:
            return f"{st.strftime('%b')} {st.year}, {st.day} {str(st.hour).zfill(2)}:{str(st.minute).zfill(2)} - {et.day} {str(et.hour).zfill(2)}:{str(et.minute).zfill(2)} UTC"
        if show_date:
            return f"{st.day}-{et.day} {st.strftime('%b')} {st.year}"

    if st.year == et.year:
        if show_time:
            return f"{st.year}, {st.strftime('%b')} {st.day} {str(st.hour).zfill(2)}:{str(st.minute).zfill(2)} - {et.strftime('%b')} {et.day} {str(et.hour).zfill(2)}:{str(et.minute).zfill(2)} UTC"
        if show_date:
            return (
                f"{st.day} {st.strftime('%b')} - {et.day} {et.strftime('%b')} {st.year}"
            )
    else:
        if show_time:
            return f"{st.day} {st.strftime('%b')} {st.year} {str(st.hour).zfill(2)}:{str(st.minute).zfill(2)} - {et.day} {et.strftime('%b')} {et.year} {str(et.hour).zfill(2)}:{str(et.minute).zfill(2)} UTC"
        if show_date:
            return f"{st.strftime('%d %b %Y')} - {et.strftime('%d %b %Y')}"
        return f"{st.strftime('%d %b %Y')} - {et.strftime('%d %b %Y')}"

    return ""


def get_time_range(
    start_time: TimestampLike,
    end_time: TimestampLike,
    freq: str | None = None,
    periods: int | None = None,
) -> pd.DatetimeIndex:
    """
    Generates a sequence of timestamps based on frequency or number of periods.

    This function either:
    - Rounds existing timestamps to the nearest interval defined by `freq`, or
    - Generates evenly spaced timestamps over a range using the specified `periods`.

    Parameters:
        freq (str, optional): A time frequency string compatible with pandas (e.g., '1H' for hourly,
            '30min' for 30-minute intervals, '1D' for daily). If provided, timestamps are rounded
            to the nearest multiple of this interval.
        periods (int, optional): The number of evenly spaced timestamps to generate. Used when
            `freq` is not provided.

    Returns:
        date_range (pandas.DatetimeIndex): A sequence of timestamps, either rounded or evenly spaced.
    """
    start_time = to_timestamp(start_time, keep_tzinfo=False)
    end_time = to_timestamp(end_time, keep_tzinfo=False)

    if not isinstance(freq, str) and not isinstance(periods, int):
        raise TypeError(
            f"{get_time_range.__name__}() missing 1 required argument: 'freq' or 'periods'"
        )
    elif isinstance(freq, str) and isinstance(periods, int):
        raise TypeError(
            f"{get_time_range.__name__}() expected 1 out of the 2 required arguments 'freq' and 'periods' but got both"
        )
    elif isinstance(freq, str):
        if freq == "MS":
            start_period = start_time.to_period("M")
            end_period = end_time.to_period("M")

            start_month = start_period.to_timestamp()
            if start_month != start_time:
                start_month = start_month + pd.Timedelta(
                    start_period.days_in_month, "D"
                )
            end_month = end_period.to_timestamp()

            n_months = int(np.round((end_month - start_month).days / 30))

            time_range = [start_month]
            for i in range(n_months):
                _period = time_range[-1].to_period("M")
                _month = time_range[-1] + pd.Timedelta(_period.days_in_month, "D")
                time_range.append(_month)
        elif freq == "SMS":
            start_period = start_time.to_period("M")
            end_period = end_time.to_period("M")

            start_month = start_period.to_timestamp()
            if start_month != start_time:
                if start_month + pd.Timedelta(14, "D") >= start_time:
                    start_month = start_month + pd.Timedelta(14, "D")
                else:
                    start_month = start_month + pd.Timedelta(
                        start_period.days_in_month, "D"
                    )
            end_month = end_period.to_timestamp()
            if end_month + pd.Timedelta(14, "D") <= end_time:
                end_month = end_month + pd.Timedelta(14, "D")

            n_semimonths = int(np.round((end_month - start_month).days / 15))

            time_range = [start_month]
            for i in range(n_semimonths):
                _ts = time_range[-1]
                if _ts.day == 15:
                    _period = time_range[-1].to_period("M")
                    _next = _period.to_timestamp() + pd.Timedelta(
                        _period.days_in_month, "D"
                    )
                else:
                    _period = time_range[-1].to_period("M")
                    _next = time_range[-1] + pd.Timedelta(14, "D")
                time_range.append(_next)
        elif "W" in freq:
            w_count = 1
            if freq[:-1] != "":
                w_count = int(freq[:-1])

            start_period = start_time.to_period("W")
            end_period = end_time.to_period("W")

            start_week = start_period.to_timestamp()
            if start_week != start_time:
                start_week = start_week + pd.Timedelta(7 * w_count, "D")
            end_week = end_period.to_timestamp()

            n_weeks = int(np.round((end_week - start_week).days / (7 * w_count)))

            time_range = [start_week]
            for i in range(n_weeks):
                _period = time_range[-1].to_period("W")
                _week = time_range[-1] + pd.Timedelta(7 * w_count, "D")
                time_range.append(_week)
        elif freq == "QS":

            def next_quarter(ts):
                q_period = pd.Timestamp(ts).to_period("Q")
                m1 = q_period.to_timestamp()
                m2 = m1 + pd.Timedelta(q_period.days_in_month, "D")
                m3 = m2 + pd.Timedelta(m2.to_period("M").days_in_month, "D")
                m4 = m3 + pd.Timedelta(m3.to_period("M").days_in_month, "D")
                return m4

            start_period = start_time.to_period("Q")
            end_period = end_time.to_period("Q")

            start_quarter = start_period.to_timestamp()
            if start_quarter != start_time:
                start_quarter = next_quarter(start_quarter)
            end_quarter = end_period.to_timestamp()

            n_quarters = int(np.round((end_quarter - start_quarter).days / 90))

            time_range = [start_quarter]
            for i in range(n_quarters):
                _quarter = next_quarter(time_range[-1])
                time_range.append(_quarter)
        elif "YS" in freq:
            y_count = 1
            if freq[:-2] != "":
                y_count = int(freq[:-2])

            start_period = start_time.to_period("Y")
            end_period = end_time.to_period("Y")

            start_year = start_period.to_timestamp()
            if start_year != start_time:
                start_year = (
                    (start_year + pd.Timedelta(366 * y_count, "D"))
                    .to_period("Y")
                    .to_timestamp()
                )
            end_year = end_period.to_timestamp()
            n_years = int(np.round((end_year - start_year).days / (365 * y_count)))

            time_range = [start_year]
            for i in range(n_years):
                _period = time_range[-1].to_period("Y")
                _year = (
                    (time_range[-1] + pd.Timedelta(366 * y_count, "D"))
                    .to_period("Y")
                    .to_timestamp()
                )
                time_range.append(_year)
        else:
            time_range = pd.date_range(
                start_time.ceil(freq), end_time.floor(freq), freq=freq
            ).to_list()
    else:
        time_range = pd.date_range(start_time, end_time, periods=periods).to_list()
    return pd.DatetimeIndex(time_range)


def lookup_value_by_timestamp(
    t: TimestampLike, times: NDArray, values: NDArray[Any]
) -> Any:
    """
    Returns the value corresponding to the timestamp closest to a given time, using interpolation.

    Parameters:
        t (TimestampLike): A single timestamp to look up.
        times (NDArray): A series of of monotonically increasing timestamps.
        values (NDArray[Any]): A series of values corresponding to each timestamp in `times`.

    Returns:
        v (Any): The value from `values` that corresponds to the closest timestamp in `times` to `t`.

    Raises:
        ValueError: If `times` and `values` have different lengths.
    """
    if t is None:
        raise ValueError(f"{lookup_value_by_timestamp.__name__}() missing `t`")
    if times is None:
        raise ValueError(f"{lookup_value_by_timestamp.__name__}() missing `times`")
    if values is None:
        raise ValueError(f"{lookup_value_by_timestamp.__name__}() missing `values`")

    t = to_timestamp(t).to_numpy()
    times = to_timestamps(times).to_numpy()
    values = np.asarray(values)

    if times.shape[0] == 0:
        raise ValueError(
            f"{lookup_value_by_timestamp.__name__}() `times` is empty but needs at least on element"
        )
    if values.shape[0] != times.shape[0]:
        raise ValueError(
            f"{lookup_value_by_timestamp.__name__}() First shapes must be the same for `values` ({values.shape[0]}) and `times` ({times.shape[0]})"
        )

    idx0 = np.searchsorted(times, t) - 1
    idx1 = np.searchsorted(times, t)

    idx0 = np.min([len(times) - 1, np.max([0, idx0])])
    idx1 = np.min([len(times) - 1, np.max([0, idx1])])

    if times[idx0] > t:
        idx0 = idx1

    total_duration = times[idx1] - times[idx0]
    duration = t - times[idx0]

    if total_duration == pd.Timedelta(0, "s"):
        frac = 0
    else:
        frac = duration / total_duration

    total_amount = values[idx1] - values[idx0]
    v = values[idx0] + total_amount * frac

    return v


@dataclass
class TimestampComparisonResult:
    is_same_year: bool
    is_same_month: bool
    is_same_day: bool
    is_same_hour: bool
    is_same_minute: bool
    is_same_second: bool

    @property
    def is_same(self) -> bool:
        return all(
            [
                self.is_same_year,
                self.is_same_month,
                self.is_same_day,
                self.is_same_hour,
                self.is_same_minute,
                self.is_same_second,
            ]
        )

    @property
    def is_same_date(self) -> bool:
        return all(
            [
                self.is_same_year,
                self.is_same_month,
                self.is_same_day,
            ]
        )

    @property
    def is_same_time(self) -> bool:
        return all(
            [
                self.is_same_hour,
                self.is_same_minute,
                self.is_same_second,
            ]
        )


def check_if_same_timestamp(
    t1: TimestampLike, t2: TimestampLike
) -> TimestampComparisonResult:
    t1 = to_timestamp(t1)
    t2 = to_timestamp(t2)

    result = TimestampComparisonResult(
        is_same_year=t1.year == t2.year,
        is_same_month=t1.month == t2.month,
        is_same_day=t1.day == t2.day,
        is_same_hour=t1.hour == t2.hour,
        is_same_minute=t1.minute == t2.minute,
        is_same_second=t1.second == t2.second,
    )

    return result


def time_to_num(
    time: NDArray | Sequence[TimestampLike],
    epoch: TimestampLike,
    format: str = "ns",
) -> NDArray:
    """
    Converts datetime-like values to numerical values relative to a given epoch.

    Args:
        time (NDArray | Iterable[TimestampLike]): Array or iterable of datetime-like values.
        epoch (TimestampLike): Reference time from which the numerical values are computed.
        format (str, optional): Time unit for conversion (e.g., "s", "ms", "us", "ns"). Defaults to "ns".

    Returns:
        NDArray: Array of floats representing the time difference from the epoch in the given unit.
    """
    time = to_timestamps(time).to_numpy()
    epoch = to_timestamp(epoch).to_numpy()
    return (time - epoch) / np.timedelta64(1, format)


def num_to_time(
    num: NDArray | Iterable,
    epoch: TimestampLike,
    format: str = "ns",
) -> NDArray:
    """
    Converts numerical time values back to datetime values relative to a given epoch.

    Args:
        num (NDArray | Iterable): Array or iterable of numeric values representing time offsets.
        epoch (TimestampLike): Reference time to which the numeric values are relative.
        format (str, optional): Time unit of the numeric values (e.g., "s", "ms", "us", "ns"). Defaults to "ns".

    Returns:
        NDArray: Array of datetime64 values corresponding to the numeric offsets from the epoch.
    """
    num = np.asarray(num)
    epoch = to_timestamp(epoch).to_numpy()
    return (num * np.timedelta64(1, format)) + epoch



from dataclasses import dataclass
from typing import Sequence

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import (
    CM_AS_INCH,
    FIGURE_HEIGHT_CURTAIN,
    FIGURE_HEIGHT_LINE,
    FIGURE_HEIGHT_SWATH,
    FIGURE_MAP_WIDTH,
    FIGURE_WIDTH_CURTAIN,
    FIGURE_WIDTH_PROFILE,
)
from ..figure_type import FigureType


@dataclass(frozen=True)
class FigureLayoutMapMainZoomProfile:
    fig: Figure
    axs_map: list[Axes]
    axs: list[Axes]
    axs_zoom: list[Axes]
    axs_profile: list[Axes]


def create_multi_figure_layout(
    rows: Sequence[FigureType | int],
    zoom_rows: Sequence[FigureType | int] | None = None,
    profile_rows: Sequence[FigureType | int] | None = None,
    map_rows: Sequence[FigureType | int] | None = None,
    wspace: float | Sequence[float] = 1.2,
    hspace: float | Sequence[float] = 1.2,
    wmain: float = FIGURE_WIDTH_CURTAIN,
    hrow: float = FIGURE_HEIGHT_CURTAIN,
    hswath: float = FIGURE_HEIGHT_SWATH,
    hline: float = FIGURE_HEIGHT_LINE,
    wprofile: float = FIGURE_WIDTH_PROFILE,
    wmap: float = FIGURE_MAP_WIDTH,
    wzoom: float = FIGURE_WIDTH_CURTAIN / 3.0,
) -> FigureLayoutMapMainZoomProfile:
    """
    Creates a complex figure layout with columns for map, main, zoom, and profile panels (in that order from left to right).

    Each panel column can have a custom sequence of figure types (e.g., row heights), and the layout
    supports both uniform and per-gap horizontal/vertical spacing.

    Args:
        main_rows (Sequence[FigureType | int]): List of figure types for the rows of the main column.
        zoom_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional zoom column.
        profile_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional profile column.
        map_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional map column.
        wspace (float | Sequence[float], optional): Horizontal spacing between columns. Can be a single value
            or list defining spacing before, between, and after columns.
        hspace (float | Sequence[float], optional): Vertical spacing between rows. Similar behavior as `wspace`.
        wmain (float, optional): Width of the main column. Default is `FIGURE_WIDTH_CURTAIN`.
        hrow (float, optional): Height of a standard row. Default is `FIGURE_HEIGHT_CURTAIN`.
        hswath (float, optional): Height of a `SwathFigure`-type row. Default is `FIGURE_HEIGHT_SWATH`.
        wprofile (float, optional): Width of the profile column.
        wmap (float, optional): Width of the map column.
        wzoom (float, optional): Width of the zoom column.

    Returns:
        tuple: A tuple containing:
            - Figure: The matplotlib figure object.
            - Sequence[Axes]: Axes for map panels (may be empty).
            - Sequence[Axes]: Axes for main panels.
            - Sequence[Axes]: Axes for zoom panels (may be empty).
            - Sequence[Axes]: Axes for profile panels (may be empty).

    Raises:
        ValueError: If the provided spacing sequences are of invalid length.
        TypeError: If spacing arguments are of unsupported types.
    """
    # Calculate number of columns
    is_map_col: bool = isinstance(map_rows, list) and len(map_rows) > 0
    is_main_col: bool = isinstance(rows, list) and len(rows) > 0
    is_zoom_col: bool = isinstance(zoom_rows, list) and len(zoom_rows) > 0
    is_profile_col: bool = isinstance(profile_rows, list) and len(profile_rows) > 0
    col_present: list[bool] = [is_map_col, is_main_col, is_zoom_col, is_profile_col]

    ncols: int = sum(col_present)

    # Calculate number of rows
    nrows_min: int = 0
    if isinstance(map_rows, list):
        for ft in map_rows:
            nrows_min += 1
            if ft == FigureType.MAP_2_ROW:
                nrows_min += 1

    nrows: int = max(nrows_min, len(rows))

    # Calulate spaces between figures
    def _calulate_spaces(
        space: float | Sequence[float],
        n: int,
        name: str,
        name_col_row: str,
    ) -> list[float]:
        if isinstance(space, Sequence):
            space = list(space)
            if len(space) < n - 1 or len(space) > n + 1:
                raise ValueError(
                    f"{name} was given as a list (size={len(space)}) and thus needs to have a size between number of {name_col_row} ({n}) -1 (i.e. only spaces between {name_col_row}) and +1 (i.e. spaces before, between and after {name_col_row})."
                )
            elif len(space) == n - 1:
                space = [0.0] + space + [0.0]
            elif len(space) == n:
                space = space + [0.0]
        elif isinstance(space, float):
            space = [0.0] + [space] * (n - 1) + [0.0]
        else:
            raise TypeError(
                f"{name} has wrong type '{type(space).__name__}'. expected types: '{float.__name__}' or ''{list.__name__}'[{float.__name__}]'"
            )
        return space

    wspace = _calulate_spaces(wspace, ncols, "wspace", "columns")
    hspace = _calulate_spaces(hspace, nrows, "hspace", "rows")

    # Calculate size ratios of figures
    def _get_ratios(
        ratios_figs: list[float],
        space: list[float],
    ) -> list[float]:
        assert len(space) == len(ratios_figs) + 1

        ratios: list[float] = []
        for i, r in enumerate(ratios_figs):
            ratios.append(space[i])
            ratios.append(r)
        ratios.append(space[-1])

        return ratios

    wratios_figs: list[float] = np.array([wmap, wmain, wzoom, wprofile])[
        col_present
    ].tolist()
    hratios_figs: list[float] = []
    for fig_type in rows:
        if isinstance(fig_type, float):
            hratios_figs.append(fig_type)
        elif fig_type == FigureType.SWATH:
            hratios_figs.append(hswath)
        elif fig_type == FigureType.LINE:
            hratios_figs.append(hline)
        elif fig_type == FigureType.CURTAIN_75:
            hratios_figs.append(hrow * 0.75)
        elif fig_type == FigureType.CURTAIN_67:
            hratios_figs.append(hrow * 0.666666667)
        elif fig_type == FigureType.CURTAIN_50:
            hratios_figs.append(hrow * 0.50)
        elif fig_type == FigureType.CURTAIN_33:
            hratios_figs.append(hrow * 0.333333333)
        elif fig_type == FigureType.CURTAIN_25:
            hratios_figs.append(hrow * 0.25)
        else:
            hratios_figs.append(hrow)
    if len(rows) < nrows_min:
        for i in range(nrows_min - len(rows)):
            hratios_figs.append(hrow)

    wratios = _get_ratios(wratios_figs, wspace)
    hratios = _get_ratios(hratios_figs, hspace)

    # Create the figure
    wfig = sum(wratios)
    hfig = sum(hratios)
    figsize = (wfig, hfig)

    fig = plt.figure(figsize=figsize)

    # Create the grid layout
    gs = gridspec.GridSpec(
        nrows=len(hratios),
        ncols=len(wratios),
        width_ratios=wratios,
        height_ratios=hratios,
        figure=fig,
        wspace=0,
        hspace=0,
        bottom=0.0,
        top=1.0,
        right=1.0,
        left=0.0,
    )

    # Create the plots
    # Create maps
    current_col: int = 1
    current_row: int = 1
    axs_map: list[Axes] = []
    axs_main: list[Axes] = []
    axs_zoom: list[Axes] = []
    axs_profile: list[Axes] = []
    ax: Axes | None
    if isinstance(map_rows, list):
        for fig_type in map_rows:
            if fig_type == FigureType.MAP_2_ROW:
                ax = fig.add_subplot(gs[current_row : current_row + 3, current_col])
                current_row += 2
            elif fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_map.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(rows, list):
        for fig_type in rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_main.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(zoom_rows, list):
        for fig_type in zoom_rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_zoom.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(profile_rows, list):
        for fig_type in profile_rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_profile.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    return FigureLayoutMapMainZoomProfile(
        fig=fig,
        axs_map=axs_map,
        axs=axs_main,
        axs_zoom=axs_zoom,
        axs_profile=axs_profile,
    )



from dataclasses import dataclass
from logging import Logger

import numpy as np

from ._exceptions import InvalidInputError
from ._parse_frame_ids import parse_frame_ids
from ._parse_orbit_and_frames import parse_orbit_and_frames
from ._parse_orbit_numbers import parse_orbit_numbers
from ._types import FrameIDStr, OrbitFrameStr, OrbitInt, _OrbitFrameInputs


def validate_combination_of_given_orbit_and_frame_range_inputs(
    start_orbit_and_frame: FrameIDStr | None,
    end_orbit_and_frame: FrameIDStr | None,
    start_orbit_number: OrbitInt | None,
    end_orbit_number: OrbitInt | None,
    orbit_numbers: list[OrbitInt] | None,
    frame_ids: list[FrameIDStr] | None,
    logger: Logger | None = None,
) -> None:
    """Raises an InvalidInputError, if combined orbit and frame options (-soaf, -eoaf) are used in combination with any other orbit (-o, -so, -eo) or frames (-f) option (exception: -oaf)."""
    try:
        if (start_orbit_and_frame is not None or end_orbit_and_frame is not None) and (
            start_orbit_number is not None
            or end_orbit_number is not None
            or orbit_numbers is not None
            or frame_ids is not None
        ):
            exception_msg = f"Options to select a range of obit and frame names (-soaf, -eoaf) can not be used in combination with the options to select only a range of orbits (-o, -so, -eo) or single frames (-f)."
            raise InvalidInputError(exception_msg)
    except InvalidInputError as e:
        if logger:
            logger.exception(e)
        raise


def parse_all_orbit_and_frame_inputs(
    args_orbit_number: list[OrbitInt] | None,
    args_start_orbit_number: OrbitInt | None,
    args_end_orbit_number: OrbitInt | None,
    args_frame_id: list[FrameIDStr] | None,
    args_orbit_and_frame: list[OrbitFrameStr] | None,
    args_start_orbit_and_frame: OrbitFrameStr | None,
    args_end_orbit_and_frame: OrbitFrameStr | None,
    logger: Logger | None = None,
) -> _OrbitFrameInputs:
    validate_combination_of_given_orbit_and_frame_range_inputs(
        start_orbit_and_frame=args_start_orbit_and_frame,
        end_orbit_and_frame=args_end_orbit_and_frame,
        start_orbit_number=args_start_orbit_number,
        end_orbit_number=args_end_orbit_number,
        orbit_numbers=args_orbit_number,
        frame_ids=args_frame_id,
        logger=logger,
    )

    full_orbits: list[OrbitInt] = []
    frame_orbits: dict[FrameIDStr, list[OrbitInt]] = {
        "A": [],
        "B": [],
        "C": [],
        "D": [],
        "E": [],
        "F": [],
        "G": [],
        "H": [],
    }
    full_orbit_range: tuple[OrbitInt | None, OrbitInt | None] = (None, None)
    frame_orbit_ranges: dict[
        FrameIDStr,
        tuple[OrbitInt | None, OrbitInt | None],
    ] = {
        "A": (None, None),
        "B": (None, None),
        "C": (None, None),
        "D": (None, None),
        "E": (None, None),
        "F": (None, None),
        "G": (None, None),
        "H": (None, None),
    }
    orbit_numbers = parse_orbit_numbers(
        args_orbit_number,
        args_start_orbit_number,
        args_end_orbit_number,
        logger=logger,
    )

    frame_ids = parse_frame_ids(args_frame_id)

    if len(frame_ids) == 0:
        full_orbits = [int(o) for o in np.unique(np.append(full_orbits, orbit_numbers.orbit_list)).flatten().tolist()]  # type: ignore
        full_orbit_range = orbit_numbers.orbit_range  # type: ignore
    else:
        for f in frame_ids:
            frame_orbits[f].extend(orbit_numbers.orbit_list)  # type: ignore
            frame_orbit_ranges[f] = orbit_numbers.orbit_range  # type: ignore

    orbit_and_frames = parse_orbit_and_frames(
        args_orbit_and_frame,
        args_start_orbit_and_frame,
        args_end_orbit_and_frame,
        logger=logger,
    )

    full_orbits = [int(o) for o in np.unique(np.append(full_orbits, orbit_and_frames.full_orbit_list)).flatten().tolist()]  # type: ignore
    if all([x is None for x in full_orbit_range]):
        full_orbit_range = orbit_and_frames.full_orbit_range  # type: ignore

    for k, v in orbit_and_frames.frame_orbits.items():  # type: ignore
        frame_orbits[k] = [int(o) for o in np.unique(np.append(frame_orbits[k], v)).flatten().tolist()]  # type: ignore

    return _OrbitFrameInputs(
        frame_orbits=frame_orbits,
        full_orbit_range=full_orbit_range,
        frame_orbit_ranges=frame_orbit_ranges,
        full_orbits=full_orbits,
        frame_ids=frame_ids,
    )



import warnings

import numpy as np
from numpy.typing import ArrayLike, NDArray


def vincenty(
    a: ArrayLike,
    b: ArrayLike,
    units: str = "km",
    tolerance: float = 1e-12,
    max_iterations: int = 10,
) -> np.float64 | NDArray[np.float64]:
    """
    Calculates the geodesic distances between points on Earth (i.e. WSG 84 ellipsoid) using Vincenty's inverse method.

    Supports single or sequences of coordiates.

    Args:
        a (ArrayLike): Coordinates [lat, lon] or array of shape (N, 2), in decimal degrees.
        b (ArrayLike): Second coordinates, same format/shape as `a`.
        units (str, optional): Output units, "km" (default) or "m".
        tolerance (float, optional): Convergence threshold in radians. Default is 1e-12.
        max_iterations (int, optional): Maximum iterations before failure. Default is 10.

    Returns:
        float or np.ndarray: The geodesic distance or distances between the point in `a` and `b`.

    Raises:
        ValueError: If input shapes are incompatible or units are invalid.

    Note:
        Uses WGS84 (a=6378137.0 m, f=1/298.257223563). May fail for nearly antipodal points.

    Examples:
        >>> geodesic([51.352757, 12.43392], [38.559, 68.856])
        4548.675334434374
        >>> geodesic([0,0], [[0,0], [10,0], [20,0]])
        array([   0.        , 1105.85483324, 2212.36625417])
        >>> geodesic([[0,0], [10,0], [20,0]], [[0,0], [10,0], [20,0]])
        array([0., 0., 0.])

    References:
        Vincenty, T. (1975). "Direct and Inverse Solutions of Geodesics on the Ellipsoid with application
        of nested equations." Survey Review, 23(176), 88-93. https://doi.org/10.1179/sre.1975.23.176.88
    """
    _a, _b = map(np.asarray, [a, b])
    coord_a, coord_b = map(np.atleast_2d, [_a, _b])
    coord_a, coord_b = map(np.radians, [coord_a, coord_b])

    if (coord_a.shape[1] != 2) or (coord_b.shape[1] != 2):
        raise ValueError(
            f"At least one passed array has a wrong shape (a={_a.shape}, b={_b.shape}). 1d arrays should be of length 2 (i.e. [lat, lon]) and 2d array should have the shape (n, 2)."
        )
    if (coord_a.shape[0] < 1) or (coord_b.shape[0] < 1):
        raise ValueError(
            f"At least one passed array contains no values (a={_a.shape}, b={_b.shape})."
        )
    if coord_a.shape[0] != coord_b.shape[0]:
        if (coord_a.shape[0] != 1) and (coord_b.shape[0] != 1):
            raise ValueError(
                f"The shapes of passed arrays dont match (a={_a.shape}, b={_b.shape}). Either both should contain the same number of coordinates or at least one of them should contain a single coordinate."
            )

    lat_1, lon_1 = coord_a[:, 0], coord_a[:, 1]
    lat_2, lon_2 = coord_b[:, 0], coord_b[:, 1]

    # WGS84 ellipsoid constants
    a = 6378137.0  # semi-major axis (equatorial radius) in meters
    f = 1 / 298.257223563  # flattening
    b = (1 - f) * a  # semi-minor axis (polar radius) in meters

    # Reduced latitudes
    beta_1 = np.arctan((1 - f) * np.tan(lat_1))
    beta_2 = np.arctan((1 - f) * np.tan(lat_2))

    initial_lon_diff = lon_2 - lon_1

    # Initialize variables for iterative solution
    lon_diff = initial_lon_diff
    sin_beta_1, cos_beta_1 = np.sin(beta_1), np.cos(beta_1)
    sin_beta_2, cos_beta_2 = np.sin(beta_2), np.cos(beta_2)
    # Track convergence for each point pair
    converged = np.full_like(lat_1, False, dtype=bool)

    for _ in range(max_iterations):
        sin_lon_diff, cos_lon_diff = np.sin(lon_diff), np.cos(lon_diff)

        sin_sigma = np.sqrt(
            (cos_beta_2 * sin_lon_diff) ** 2
            + (cos_beta_1 * sin_beta_2 - sin_beta_1 * cos_beta_2 * cos_lon_diff) ** 2
        )
        cos_sigma = (sin_beta_1 * sin_beta_2) + (cos_beta_1 * cos_beta_2 * cos_lon_diff)
        sigma = np.arctan2(sin_sigma, cos_sigma)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            sin_alpha = cos_beta_1 * cos_beta_2 * sin_lon_diff / sin_sigma
        sin_alpha = np.nan_to_num(sin_alpha, nan=0.0)
        cos2_alpha = 1 - sin_alpha**2

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            cos2_sigma_m = np.where(
                cos2_alpha != 0.0,
                cos_sigma - ((2 * sin_beta_1 * sin_beta_2) / cos2_alpha),
                0.0,
            )
        cos2_sigma_m = np.nan_to_num(cos2_sigma_m, nan=0.0)

        C = f / 16 * cos2_alpha * (4 + f * (4 - 3 * cos2_alpha))

        previous_lon_diff = lon_diff
        lon_diff = initial_lon_diff + (1 - C) * f * sin_alpha * (
            sigma
            + C
            * sin_sigma
            * (cos2_sigma_m + C * cos_sigma * (-1 + 2 * cos2_sigma_m**2))
        )
        converged = converged | (np.abs(lon_diff - previous_lon_diff) < tolerance)
        if np.all(converged):
            break

    u2 = cos2_alpha * (a**2 - b**2) / b**2
    A = 1 + u2 / 16384 * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
    B = u2 / 1024 * (256 + u2 * (-128 + u2 * (74 - 47 * u2)))

    delta_sigma = (
        B
        * sin_sigma
        * (
            cos2_sigma_m
            + B
            / 4
            * (
                cos_sigma * (-1 + 2 * cos2_sigma_m**2)
                - B
                / 6
                * cos2_sigma_m
                * (-3 + 4 * sin_sigma**2)
                * (-3 + 4 * cos2_sigma_m**2)
            )
        )
    )

    distance = b * A * (sigma - delta_sigma)

    if units == "km":
        distance = distance / 1000.0
    elif units != "m":
        raise ValueError(
            f"{vincenty.__name__}() Invalid units : {units}. Use 'm' or 'km' instead."
        )

    if len(_a.shape) == 1 and len(_b.shape) == 1:
        return distance[0]

    return distance



from typing import TYPE_CHECKING, Union

import xarray as xr

from .fill_values import _convert_all_fill_values_to_nan

# To enable static type checking while avoiding circular import error FileAgency
# is imported like this, so it can be used as a string-based type hin later.
if TYPE_CHECKING:
    from .file_info.agency import FileAgency


def read_science_data(
    filepath: str,
    agency: Union["FileAgency", None] = None,
    ensure_nans: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Opens the science data of a EarthCARE file as a `xarray.Dataset`."""
    from .file_info.agency import (
        FileAgency,  # Imported inside function to avoid circular import error
    )

    if agency is None:
        agency = FileAgency.from_input(filepath)

    if agency == FileAgency.ESA:
        ds = xr.open_dataset(filepath, group="ScienceData", engine="h5netcdf", **kwargs)
    elif agency == FileAgency.JAXA:
        df_cpr_geo = xr.open_dataset(
            filepath, group="ScienceData/Geo", engine="h5netcdf", phony_dims="sort"
        )
        df_cpr_data = xr.open_dataset(
            filepath, group="ScienceData/Data", engine="h5netcdf", phony_dims="sort"
        )
        ds = xr.merge([df_cpr_data, df_cpr_geo])
        ds.encoding["source"] = df_cpr_data.encoding["source"]
    else:
        raise NotImplementedError()

    if ensure_nans:
        ds = _convert_all_fill_values_to_nan(ds)

    return ds



from typing import Literal

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ._vincenty import vincenty as geodesic


def get_cumulative_distances(
    lats: ArrayLike, lons: ArrayLike, units: Literal["m", "km"] = "m"
) -> NDArray:
    lats = np.array(lats)
    lons = np.array(lons)
    if lats.shape != lons.shape:
        raise ValueError(
            f"Shape mismatch: 'lats' {lats.shape} and 'lons' {lons.shape} must have the same shape"
        )
    if lats.shape[0] == 1:
        return np.array([0])
    elif lats.shape[0] == 0:
        return np.array([])
    coords = np.vstack((lats, lons)).T
    distances = geodesic(coords[0:-1], coords[1::], units=units)
    distances = np.append(0, distances)
    cumulative_distances = np.cumsum(distances)
    return cumulative_distances



import numpy as np
from xarray import DataArray, Dataset

from ...constants import ALONG_TRACK_DIM, EC_LATITUDE_FRAME_BOUNDS, TRACK_LAT_VAR
from ...xarray_utils import insert_var
from .header_group import read_header_data


def get_frame_id(ds: Dataset) -> str:
    if "frameID" in ds:
        return str(ds.frameID.values)
    return str(read_header_data(ds).frameID.values.astype(str))


def get_frame_along_track(
    ds: Dataset,
    lat_var: str = TRACK_LAT_VAR,
    frame_id: str | None = None,
) -> tuple[int, int]:
    if not isinstance(frame_id, str):
        frame_id = get_frame_id(ds)
    lat_framestart, lat_framestop = EC_LATITUDE_FRAME_BOUNDS[frame_id]

    lat = ds[lat_var].data

    if lat_framestart == lat_framestop:
        if lat_framestart > 0:
            idxs = np.argwhere(lat >= lat_framestart)
        else:
            idxs = np.argwhere(lat <= lat_framestart)
    elif lat_framestart < lat_framestop:
        idxs = np.argwhere(np.logical_and(lat >= lat_framestart, lat <= lat_framestop))
    else:
        idxs = np.argwhere(np.logical_and(lat <= lat_framestart, lat >= lat_framestop))

    slice_tuple = int(idxs[0][0]), int(idxs[-1][0]) + 1

    return slice_tuple


def trim_to_latitude_frame_bounds(
    ds: Dataset,
    along_track_dim: str = ALONG_TRACK_DIM,
    lat_var: str = TRACK_LAT_VAR,
    frame_id: str | None = None,
    add_trim_index_offset_var: bool = True,
    trim_index_offset_var_name: str = "trim_index_offset",
) -> Dataset:
    """
    Trims the dataset to the region within the latitude frame bounds.

    Args:
        ds (xarray.Dataset):
            Input dataset to be trimmed.
        along_track_dim (str, optional):
            Dimension along which to trim. Defaults to ALONG_TRACK_DIM.
        lat_var (str, optional):
            Name of the latitude variable. Defaults to TRACK_LAT_VAR.
        frame_id (str | None, optional):
            EarthCARE frame ID (single character between "A" and "H").
            If given, speeds up trimming. Defaults to None.
        add_trim_index_offset_var (bool, optional):
            Whether the index offset between the original and trimmed dataset is stored
            in the trimmed dataset (variable: "trim_index_offset"). Defaults to True.

    Returns:
        xarray.Dataset: Trimmed dataset.
    """
    slice_tuple = get_frame_along_track(
        ds,
        lat_var=lat_var,
        frame_id=frame_id,
    )
    ds = ds.isel({along_track_dim: slice(*slice_tuple)})
    if add_trim_index_offset_var:
        ds = insert_var(
            ds=ds,
            var=trim_index_offset_var_name,
            data=int(slice_tuple[0]),
            index=0,
            after_var="processing_start_time",
        )
        ds[trim_index_offset_var_name] = ds[trim_index_offset_var_name].assign_attrs(
            {
                "earthcarekit": "Added by earthcarekit: Used to calculate the index in the original, untrimmed dataset, i.e. by addition."
            }
        )
    return ds



import numpy as np


class EmptyFilterResultError(Exception):
    min_distance: float

    def __init__(self, *args: object, min_distance: int | float | np.number) -> None:
        if not isinstance(min_distance, (int | float)):
            raise TypeError(
                f"`min_distance` must be numerical, but got type '{type(min_distance).__name__}'"
            )
        super().__init__(*args, f"Min. distance = {min_distance}")
        self.min_distance = float(min_distance)



import numpy as np
import xarray as xr
from numpy.typing import ArrayLike, NDArray

from ..constants import TRACK_LAT_VAR, TRACK_LON_VAR
from ..np_array_utils import flatten_array


def get_coords(
    ds: xr.Dataset,
    *,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    flatten: bool = False,
) -> NDArray:
    """Takes a `xarray.Dataset` and returns the lat/lon coordinates as a numpy array.

    Args:
        lat_var (str, optional): Name of the latitude variable. Defaults to TRACK_LAT_VAR.
        lon_var (str, optional): Name of the longitude variable. Defaults to TRACK_LON_VAR.
        flatten (bool, optional):
            If True, the coordinates will be flattened to a 2D array

            - 1st dimension: time
            - 2nd dimension: lat/lon

    Returns:
        numpy.array: The extracted lat/lon coordinates.
    """
    lat = ds[lat_var].values
    lon = ds[lon_var].values
    coords = np.stack((lat, lon)).transpose()

    if len(coords.shape) > 2 and flatten:
        coords = coords.reshape(-1, 2)
    return coords


def get_central_coords(
    latitude: ArrayLike,
    longitude: ArrayLike,
) -> tuple[float, float]:
    """Calculates the central lat/lon coordinates."""
    from .convertsions import ecef_to_geo, geo_to_ecef

    lats: NDArray = flatten_array(latitude)
    lons: NDArray = flatten_array(longitude)

    coords_ecef = np.array([geo_to_ecef(lat=lt, lon=ln) for lt, ln in zip(lats, lons)])
    coords_ecef_min = np.nanmin(coords_ecef, axis=0)
    coords_ecef_max = np.nanmax(coords_ecef, axis=0)
    coords_ecef_central = (coords_ecef_min + coords_ecef_max) * 0.5
    coords_geo_central = ecef_to_geo(
        coords_ecef_central[0], coords_ecef_central[1], coords_ecef_central[2]
    )
    return (coords_geo_central[0], coords_geo_central[1])


def get_central_latitude(
    latitude: ArrayLike,
) -> float:
    """Calculates the central latitude coordinate."""
    lats: NDArray = flatten_array(latitude)
    lons: NDArray = np.zeros(lats.shape)

    central_coords = get_central_coords(lats, lons)

    return central_coords[0]


def get_central_longitude(
    longitude: ArrayLike,
) -> float:
    """Calculates the central longitude coordinate."""
    lons: NDArray = flatten_array(longitude)
    lats: NDArray = np.zeros(lons.shape)

    central_coords = get_central_coords(lats, lons)

    return central_coords[1]



from typing import Iterable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray

from ..constants import ALONG_TRACK_DIM, TIME_VAR, TRACK_LAT_VAR
from ..np_array_utils import pad_true_sequence
from ..time import TimeRangeLike, TimestampLike, to_timestamp
from ..typing import NumericPairNoneLike, validate_numeric_pair


def _get_pole_crossing_masks(
    ds: xr.Dataset,
    lat_var: str = TRACK_LAT_VAR,
) -> tuple[bool, bool, NDArray[np.bool_], NDArray[np.bool_]]:
    lats: NDArray = ds[lat_var].values
    lats_diff: NDArray = np.diff(lats)

    satellite_crosses_pole: bool = (lats_diff[0] > 0) != (lats_diff[-1] > 0)

    is_first_increase: bool = lats_diff[0] > 0

    mask_before_pole: NDArray[np.bool_]
    mask_after_pole: NDArray[np.bool_]
    if is_first_increase:
        mask_before_pole = np.append(lats_diff[0], lats_diff) > 0
        mask_after_pole = np.logical_not(mask_before_pole)
    else:
        mask_before_pole = np.append(lats_diff[0], lats_diff) <= 0
        mask_after_pole = np.logical_not(mask_before_pole)

    return satellite_crosses_pole, is_first_increase, mask_before_pole, mask_after_pole


def filter_latitude(
    ds: xr.Dataset,
    lat_range: NumericPairNoneLike,
    start_before_pole: bool = True,
    end_before_pole: bool = True,
    only_center: bool = False,
    lat_var: str = TRACK_LAT_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    pad_idxs: int = 0,
) -> xr.Dataset:
    """
    Filters a dataset to include only points within a specified latitude range.

    Args:
        ds (xr.Dataset): Input dataset with geolocation data.
        lat_range (NumericPairNoneLike): A pair of latitude values (min_lat, max_lat) defining the selection range.
        start_before_pole (bool, optional): If True, selection starts before the pole when the track crosses one. Defaults to True.
        end_before_pole (bool, optional): If True, selection ends before the pole when the track crosses one. Defaults to True.
        only_center (bool, optional): If True, only the sample at the center index of selection is returned. Defaults to False.
        lat_var (str, optional): Name of the latitude variable. Defaults to TRACK_LAT_VAR.
        along_track_dim (str, optional): Dimension along which to apply filtering. Defaults to ALONG_TRACK_DIM.
        pad_idxs (int, optional): Number of additional samples added at both sides of the selection. Defaults to 0.

    Raises:
        ValueError: If selection is empty.

    Returns:
        xr.Dataset: Filtered dataset containing only points within the specified latitude range.
    """
    lats = ds[lat_var].values

    satellite_crosses_pole, is_first_increase, mask_before_pole, mask_after_pole = (
        _get_pole_crossing_masks(ds, lat_var=lat_var)
    )

    lat_range = validate_numeric_pair(lat_range, fallback=(lats[0], lats[-1]))

    lats_mask: NDArray[np.bool_] = (lats >= np.min(lat_range)) & (
        lats <= np.max(lat_range)
    )

    if satellite_crosses_pole and start_before_pole and not end_before_pole:
        if is_first_increase:
            mask_from_start = lats >= lat_range[0]
            mask_from_end = lats >= lat_range[1]
        else:
            mask_from_start = lats <= lat_range[0]
            mask_from_end = lats <= lat_range[1]

        mask_from_start_before_pole = np.logical_and(mask_before_pole, mask_from_start)
        mask_from_end_after_pole = np.logical_and(mask_after_pole, mask_from_end)

        mask = np.logical_or(mask_from_start_before_pole, mask_from_end_after_pole)
    elif satellite_crosses_pole and start_before_pole and end_before_pole:
        mask = np.logical_and(lats_mask, mask_before_pole)
    elif satellite_crosses_pole and not start_before_pole:
        mask = np.logical_and(lats_mask, mask_after_pole)
    else:
        mask = lats_mask

    if only_center:
        mask_true_idxs = np.where(mask)[0]
        if len(mask_true_idxs) > 0:
            idx_center = mask_true_idxs[len(mask_true_idxs) // 2]
            mask[:] = False
            mask[idx_center] = True

    mask = pad_true_sequence(mask, pad_idxs)

    if np.sum(mask) == 0:
        msg = f"No data falls into the given latitude range!\nIn the dataset latitude falls between {np.min(lats)} and {np.max(lats)}.\n"
        if satellite_crosses_pole:
            msg += "Note that the satellite crosses a pole (set `start_before_pole` and `end_before_pole`\nto clarify how the start and end of the range should be interpreted)."
        else:
            msg += "The satellite is not crossing a pole."
        raise ValueError(msg)

    da_mask: xr.DataArray = xr.DataArray(mask, dims=[along_track_dim], name=lat_var)

    ds_new: xr.Dataset = xr.Dataset(
        {
            var: (
                ds[var].copy().where(da_mask, drop=True)
                if along_track_dim in ds[var].dims
                else ds[var].copy()
            )
            for var in ds.data_vars
        }
    )
    ds_new.attrs = ds.attrs.copy()
    ds_new.encoding = ds.encoding.copy()

    return ds_new
